# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit
# for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

# This module provides the functions required to process the
# standard burnman formula formats.
# tools.chemistry returns the number of atoms and molar mass of a compound
# given its unit formula as an argument.
# process_solution_chemistry returns information required to calculate
# solution properties from a set of endmember formulae

from __future__ import absolute_import
import re
import numpy as np
from fractions import Fraction
from collections import Counter
import pkgutil
from string import ascii_uppercase as ucase
from sympy import nsimplify


def read_masses():
    """
    A simple function to read a file with a two column list of
    elements and their masses into a dictionary
    """
    datastream = pkgutil.get_data("burnman", "data/input_masses/atomic_masses.dat")
    datalines = [
        line.strip() for line in datastream.decode("ascii").split("\n") if line.strip()
    ]
    lookup = dict()
    for line in datalines:
        data = "%".join(line.split("%")[:1]).split()
        if data != []:
            lookup[data[0]] = float(data[1])
    return lookup


"""
atomic_masses is a dictionary of atomic masses
"""
atomic_masses = read_masses()

"""
IUPAC_element_order provides a list of all the elements.
Element order is based loosely on electronegativity,
following the scheme suggested by IUPAC, except that H
comes after the Group 16 elements, not before them.
"""
IUPAC_element_order = [
    "v",
    "Og",
    "Rn",
    "Xe",
    "Kr",
    "Ar",
    "Ne",
    "He",  # Group 18
    "Fr",
    "Cs",
    "Rb",
    "K",
    "Na",
    "Li",  # Group 1 (not H)
    "Ra",
    "Ba",
    "Sr",
    "Ca",
    "Mg",
    "Be",  # Group 2
    "Lr",
    "No",
    "Md",
    "Fm",
    "Es",
    "Cf",
    "Bk",
    "Cm",
    "Am",
    "Pu",
    "Np",
    "U",
    "Pa",
    "Th",
    "Ac",  # Actinides
    "Lu",
    "Yb",
    "Tm",
    "Er",
    "Ho",
    "Dy",
    "Tb",
    "Gd",
    "Eu",
    "Sm",
    "Pm",
    "Nd",
    "Pr",
    "Ce",
    "La",  # Lanthanides
    "Y",
    "Sc",  # Group 3
    "Rf",
    "Hf",
    "Zr",
    "Ti",  # Group 4
    "Db",
    "Ta",
    "Nb",
    "V",  # Group 5
    "Sg",
    "W",
    "Mo",
    "Cr",  # Group 6
    "Bh",
    "Re",
    "Tc",
    "Mn",  # Group 7
    "Hs",
    "Os",
    "Ru",
    "Fe",  # Group 8
    "Mt",
    "Ir",
    "Rh",
    "Co",  # Group 9
    "Ds",
    "Pt",
    "Pd",
    "Ni",  # Group 10
    "Rg",
    "Au",
    "Ag",
    "Cu",  # Group 11
    "Cn",
    "Hg",
    "Cd",
    "Zn",  # Group 12
    "Nh",
    "Tl",
    "In",
    "Ga",
    "Al",
    "B",  # Group 13
    "Fl",
    "Pb",
    "Sn",
    "Ge",
    "Si",
    "C",  # Group 14
    "Mc",
    "Bi",
    "Sb",
    "As",
    "P",
    "N",  # Group 15
    "Lv",
    "Po",
    "Te",
    "Se",
    "S",
    "O",  # Group 16
    "H",  # hydrogen
    "Ts",
    "At",
    "I",
    "Br",
    "Cl",
    "F",
]  # Group 17


def dictionarize_formula(formula):
    """
    A function to read a chemical formula string and
    convert it into a dictionary

    :param formula: Chemical formula, written in the XnYm format, where
        the formula has n atoms of element X and m atoms of element Y
    :type formula: str

    :returns: The same chemical formula, but expressed as a dictionary.
    :rtype: dict
    """
    f = dict()
    elements = re.findall("[A-Z][^A-Z]*", formula)
    for element in elements:
        element_name = re.split("[0-9][^A-Z]*", element)[0]
        element_atoms = re.findall("[0-9][^A-Z]*", element)
        if len(element_atoms) == 0:
            element_atoms = Fraction(1.0)
        else:
            element_atoms = Fraction(element_atoms[0])
        f[element_name] = f.get(element_name, 0.0) + element_atoms

    return f


def sum_formulae(formulae, amounts=None):
    """
    Adds together a set of formulae.

    :param formulae: List of chemical formulae.
    :type formulae: list of dictionary or counter objects
    :param amounts: List of amounts of each formula.
    :type amounts: list of floats

    :returns: The sum of the user-provided formulae
    :rtype: Counter object
    """
    if amounts is None:
        amounts = [1.0 for formula in formulae]
    else:
        assert len(formulae) == len(amounts)

    summed_formula = Counter()
    for i, formula in enumerate(formulae):
        summed_formula.update(
            Counter(
                {
                    element: amounts[i] * n_atoms
                    for (element, n_atoms) in formula.items()
                }
            )
        )
    return summed_formula


def formula_mass(formula):
    """
    A function to take a chemical formula and compute the formula mass.

    :param formula: A chemical formula
    :type formula: dict or Counter object

    :returns: The mass per mole of formula [kg]
    :rtype: float
    """
    mass = sum(formula[element] * atomic_masses[element] for element in formula)
    return mass


def convert_formula(formula, to_type="mass", normalize=False):
    """
    Converts a chemical formula from one type (mass or molar)
    into the other. Renormalises amounts if normalize=True.

    :param formula: A chemical formula.
    :type formula: dict or Counter object

    :param to_type: Conversion type, one of 'mass' or 'molar'.
    :type to_type: str

    :param normalize: Whether or not to normalize the converted formula to 1.
    :type normalize: bool

    :returns: The converted formula.
    :rtype: dict
    """

    if to_type == "mass":
        f = {
            element: n_atoms * atomic_masses[element]
            for (element, n_atoms) in formula.items()
        }
    elif to_type == "molar":
        f = {
            element: n_atoms / atomic_masses[element]
            for (element, n_atoms) in formula.items()
        }
    else:
        raise Exception(
            "Value of parameter to_type not recognised. "
            'Should be either "mass" or "molar".'
        )

    if normalize:
        s = np.sum([n for (element, n) in f.items()])
        f = {element: n / s for (element, n) in f.items()}

    return f


def process_solution_chemistry(solution_model):
    """
    This function parses a class instance with a "formulas"
    attribute containing site information, e.g.

        [ '[Mg]3[Al]2Si3O12', '[Mg]3[Mg1/2Si1/2]2Si3O12' ]

    It outputs the bulk composition of each endmember
    (removing the site information), and also a set of
    variables and arrays which contain the site information.
    These are output in a format that can easily be used to
    calculate activities and gibbs free energies, given
    molar fractions of the phases and pressure
    and temperature where necessary.

    :param solution_model: Class must have a "formulas" attribute,
        containing a list of chemical formulae with site information
    :type solution model: instance of class

    :rtype: None

    .. note:: Nothing is returned from this function, but the solution_model
        object gains the following attributes:

        * solution_formulae [list of dictionaries]
            List of endmember formulae in dictionary form.
        * empty_formula [string]
            Abbreviated chemical formula with sites denoted by empty
            square brackets.
        * general_formula [string]
            General chemical formula with sites denoted by
            square brackets filled with a comma-separated list of species
        * n_sites [integer]
            Number of sites in the solution.
            Should be the same for all endmembers.
        * sites [list of lists of strings]
            A list of species for each site in the solution.
        * site_names [list of strings]
            A list of species_site pairs in the solution, where
            each distinct site is given by a unique uppercase letter
            e.g. ['Mg_A', 'Fe_A', 'Al_A', 'Al_B', 'Si_B'].
        * n_occupancies [integer]
            Sum of the number of possible species on each of the sites
            in the solution.
            Example: A binary solution [[A][B],[B][C1/2D1/2]] would have
            n_occupancies = 5, with two possible species on
            Site 1 and three on Site 2.
        * site_multiplicities [2D array of floats]
            A 1D array for each endmember in the solution,
            containing the multiplicities of each site per formula unit.
            To simplify computations later, the multiplicities
            are repeated for each species on each site, so the shape of
            this attribute is (n_endmembers, n_site_species).
        * endmember_occupancies [2d array of floats]
            A 1D array for each endmember in the solution,
            containing the fraction of atoms of each species on each site.
        * endmember_noccupancies [2d array of floats]
            A 1D array for each endmember in the solution,
            containing the number of atoms of each species on each site
            per mole of endmember.
    """
    formulae = solution_model.formulas
    n_sites = formulae[0].count("[")
    n_endmembers = len(formulae)

    # Check the number of sites is the same for all endmembers
    if not np.all(np.array([f.count("[") for f in formulae]) == n_sites):
        raise Exception("All formulae must have the same " "number of distinct sites.")

    solution_formulae = [{} for i in range(n_endmembers)]
    sites = [[] for i in range(n_sites)]
    list_occupancies = []
    list_multiplicities = np.empty(shape=(n_endmembers, n_sites))
    n_occupancies = 0

    # Number of unique site occupancies (e.g.. Mg on X etc.)
    for i_mbr in range(n_endmembers):
        list_occupancies.append([[0] * len(sites[site]) for site in range(n_sites)])
        s = re.split(r"\[", formulae[i_mbr])[1:]

        for i_site, site_string in enumerate(s):
            site_split = re.split(r"\]", site_string)
            site_occupancy = site_split[0]

            mult = re.split("[A-Z][^A-Z]*", site_split[1])[0]
            if mult == "":
                list_multiplicities[i_mbr][i_site] = Fraction(1.0)
            else:
                list_multiplicities[i_mbr][i_site] = Fraction(mult)

            # Loop over species on a site
            species = re.findall("[A-Z][^A-Z]*", site_occupancy)

            for sp in species:
                # Find the species and its proportion on the site
                species_split = re.split("([0-9][^A-Z]*)", sp)
                name_of_species = species_split[0]
                if len(species_split) == 1:
                    proportion_species_on_site = Fraction(1.0)
                else:
                    proportion_species_on_site = Fraction(species_split[1])

                solution_formulae[i_mbr][name_of_species] = solution_formulae[
                    i_mbr
                ].get(name_of_species, 0.0) + (
                    list_multiplicities[i_mbr][i_site] * proportion_species_on_site
                )

                if name_of_species not in sites[i_site]:
                    n_occupancies += 1
                    sites[i_site].append(name_of_species)
                    i_el = sites[i_site].index(name_of_species)
                    for parsed_mbr in range(len(list_occupancies)):
                        list_occupancies[parsed_mbr][i_site].append(0)
                else:
                    i_el = sites[i_site].index(name_of_species)
                list_occupancies[i_mbr][i_site][i_el] = proportion_species_on_site

            # Loop over species after site
            if len(site_split) != 1:
                not_in_site = str(filter(None, site_split[1]))
                not_in_site = not_in_site.replace(mult, "", 1)
                for enamenumber in re.findall("[A-Z][^A-Z]*", not_in_site):
                    sp = list(filter(None, re.split(r"(\d+)", enamenumber)))
                    # Look up number of atoms of element
                    if len(sp) == 1:
                        nel = 1.0
                    else:
                        nel = float(float(sp[1]))
                    solution_formulae[i_mbr][sp[0]] = (
                        solution_formulae[i_mbr].get(sp[0], 0.0) + nel
                    )

    # Site occupancies and multiplicities
    endmember_occupancies = np.empty(shape=(n_endmembers, n_occupancies))
    site_multiplicities = np.empty(shape=(n_endmembers, n_occupancies))

    for i_mbr in range(n_endmembers):
        n_species = 0
        for i_site in range(n_sites):
            for i_el in range(len(list_occupancies[i_mbr][i_site])):
                endmember_occupancies[i_mbr][n_species] = list_occupancies[i_mbr][
                    i_site
                ][i_el]
                site_multiplicities[i_mbr][n_species] = list_multiplicities[i_mbr][
                    i_site
                ]
                n_species += 1

    # Site names
    solution_model.site_names = []
    for i, species in enumerate(sites):
        for sp in species:
            solution_model.site_names.append("{0}_{1}".format(sp, ucase[i]))

    # Finally, make attributes for solution model instance:
    solution_model.solution_formulae = solution_formulae
    solution_model.n_sites = n_sites
    solution_model.sites = sites
    solution_model.site_multiplicities = site_multiplicities
    solution_model.n_occupancies = n_occupancies
    solution_model.endmember_occupancies = endmember_occupancies
    solution_model.endmember_noccupancies = np.einsum(
        "ij, ij->ij", endmember_occupancies, site_multiplicities
    )

    solution_model.empty_formula = re.sub(
        "([\\[]).*?([\\]])", "\\g<1>\\g<2>", solution_model.formulas[0]
    )
    split_empty = solution_model.empty_formula.split("[")
    solution_model.general_formula = split_empty[0]
    for i in range(n_sites):
        solution_model.general_formula += f"[{','.join(sites[i])}{split_empty[i+1]}"


def site_occupancies_to_strings(
    site_species_names, site_multiplicities, endmember_occupancies
):
    """
    Converts a list of endmember site occupancies into a list
    of string representations of those occupancies.

    :param site_species_names: A list of list of strings,
        giving the names of the species which reside on each site.
        List of sites, each of which contains a list of the species
        occupying each site.
    :type site_species_names: 2D list of strings

    :param site_multiplicities: List of floats giving the multiplicity
        of each site. If 2D, must have the same shape as endmember_occupancies.
        If 1D, must be either the same length as the number of sites, or
        the same length as site_species_names
        (with an implied repetition of the same
        number for each species on a given site).
    :type site_multiplicities: 1D or 2D numpy array of floats

    :param endmember_occupancies: A list of site-species occupancies
        for each endmember. The first dimension loops over the endmembers, and the
        second dimension loops over the site-species occupancies for that endmember.
        The total number and order of occupancies must
        be the same as the strings in site_species_names.
    :type endmember_occupancies: 2D numpy array of floats

    :returns: A list of strings in standard burnman format.
        For example, [Mg]3[Al]2 would correspond to the
        classic two-site pyrope garnet.
    :rtype: list of strings
    """

    site_multiplicities = np.array(site_multiplicities)
    endmember_occupancies = np.array(endmember_occupancies)
    n_endmembers = endmember_occupancies.shape[0]

    if len(site_multiplicities.shape) == 1:
        # Site multiplicities should either be given on a per-site basis,
        # or a per-species basis
        if len(site_species_names) == len(site_multiplicities):
            site_mults = []

            for i, site in enumerate(site_species_names):
                for species in site:
                    site_mults.append(site_multiplicities[i])

            site_multiplicities = np.array(site_mults)

        elif len(endmember_occupancies[0]) != len(site_multiplicities):
            raise Exception(
                "Site multiplicities should either be given "
                "on a per-site basis or a per-species basis"
            )

        site_multiplicities = np.einsum(
            "i, j->ij", np.ones(n_endmembers), site_multiplicities
        )
    elif len(site_multiplicities.shape) == 2:
        if site_multiplicities.shape != endmember_occupancies.shape:
            raise Exception(
                "If site_multiplicities is 2D, it should have "
                "the same shape as endmember_occupancies. "
                "They currently have shapes "
                f"{site_multiplicities.shape} and "
                f"{endmember_occupancies.shape}."
            )
    else:
        raise Exception("Site multiplicities should either be 1D or 2D.")

    site_formulae = []
    for i_mbr, mbr_occupancies in enumerate(endmember_occupancies):
        i = 0
        site_formulae.append("")
        for site in site_species_names:
            amounts = mbr_occupancies[i : i + len(site)]
            mult = site_multiplicities[i_mbr, i]
            if np.abs(mult - 1.0) < 1.0e-12:
                mult = ""
            else:
                mult = str(nsimplify(mult))
            amounts /= sum(amounts)
            site_occupancy = formula_to_string(dict(zip(site, amounts)))
            site_formulae[-1] += "[{0}]{1}".format(site_occupancy, mult)
            i += len(site)

    return site_formulae


def compositional_array(formulae):
    """
    :param formulae: List of chemical formulae
    :type formulae: list of dicts

    :returns: Array of endmember formulae and a list of elements.
    :rtype: 2D numpy.array of floats and a list of strs
    """
    elements = []
    for formula in formulae:
        for element in formula:
            if element not in elements:
                elements.append(element)

    formula_array = ordered_compositional_array(formulae, elements)

    return formula_array, elements


def ordered_compositional_array(formulae, elements):
    """
    :param formulae: List of chemical formulae
    :type formulae: list of dicts

    :param elements : List of elements
    :type elements: list of strings

    :returns: Array of endmember formulae
    :rtype: 2D array of floats
    """
    formula_array = np.zeros(shape=(len(formulae), len(elements)))
    for idx, formula in enumerate(formulae):
        for element in formula:
            assert element in elements
            formula_array[idx][elements.index(element)] = formula[element]

    return formula_array


def formula_to_string(formula):
    """
    :param formula: Chemical formula
    :type formula: dict or Counter

    :returns: A formula string, with element order as given in the list
        IUPAC_element_order.
        If one or more keys in the dictionary are not one of the elements
        in the periodic table, then they are added at the end of the string.
    :rtype: str
    """

    formula_string = ""
    for e in IUPAC_element_order:
        if e in formula and np.abs(formula[e]) > 1.0e-12:
            if np.abs(formula[e] - 1.0) < 1.0e-12:
                formula_string += e
            else:
                formula_string += e + str(nsimplify(formula[e]))

    for e in formula:
        if e not in IUPAC_element_order:
            if e in formula and np.abs(formula[e]) > 1.0e-12:
                if np.abs(formula[e] - 1.0) < 1.0e-12:
                    formula_string += e
                else:
                    formula_string += e + str(nsimplify(formula[e]))

    return formula_string


def sort_element_list_to_IUPAC_order(element_list):
    """
    :param element_list : List of elements.
    :type element_list: list

    :returns: List of elements sorted into IUPAC order
    :rtype: list
    """
    sorted_list = [e for e in IUPAC_element_order if e in element_list]
    assert len(sorted_list) == len(element_list)
    return sorted_list


def convert_fractions(composite, phase_fractions, input_type, output_type):
    """
    Takes a composite with a set of user defined molar, volume
    or mass fractions (which do not have to be the fractions
    currently associated with the composite) and
    converts the fractions to molar, mass or volume.

    Conversions to and from mass require a molar mass to be
    defined for all phases. Conversions to and from volume
    require set_state to have been called for the composite.

    :param composite: Composite for which fractions are to be defined.
    :type composite: :class:`~burnman.Composite`

    :param phase_fractions: List of input phase fractions
        (of type input_type).
    :type phase_fractions: list of floats

    :param input_type: Input fraction type. One of 'molar', 'mass' or 'volume'.
    :type input_type: str

    :param output_type: Output fraction type. One of 'molar', 'mass' or 'volume'.
    :type output_type: str

    :returns: List of output phase fractions (of type output_type)
    :rtype: list of floats
    """
    if input_type == "volume" or output_type == "volume":
        if composite.temperature is None:
            raise Exception(
                composite.to_string()
                + ".set_state(P, T) has not been called, so volume fractions are currently undefined. Exiting."
            )

    if input_type == "molar":
        molar_fractions = phase_fractions
    if input_type == "volume":
        total_moles = sum(
            volume_fraction / phase.molar_volume
            for volume_fraction, phase in zip(phase_fractions, composite.phases)
        )
        molar_fractions = [
            volume_fraction / (phase.molar_volume * total_moles)
            for volume_fraction, phase in zip(phase_fractions, composite.phases)
        ]
    if input_type == "mass":
        total_moles = sum(
            mass_fraction / phase.molar_mass
            for mass_fraction, phase in zip(phase_fractions, composite.phases)
        )
        molar_fractions = [
            mass_fraction / (phase.molar_mass * total_moles)
            for mass_fraction, phase in zip(phase_fractions, composite.phases)
        ]

    if output_type == "volume":
        total_volume = sum(
            molar_fraction * phase.molar_volume
            for molar_fraction, phase in zip(molar_fractions, composite.phases)
        )
        output_fractions = [
            molar_fraction * phase.molar_volume / total_volume
            for molar_fraction, phase in zip(molar_fractions, composite.phases)
        ]
    elif output_type == "mass":
        total_mass = sum(
            molar_fraction * phase.molar_mass
            for molar_fraction, phase in zip(molar_fractions, composite.phases)
        )
        output_fractions = [
            molar_fraction * phase.molar_mass / total_mass
            for molar_fraction, phase in zip(molar_fractions, composite.phases)
        ]
    elif output_type == "molar":
        output_fractions = molar_fractions

    return output_fractions


def reaction_matrix_as_strings(reaction_matrix, compound_names):
    """
    Returns a list of string representations of all the reactions in
    reaction_matrix.

    :param reaction_matrix: Matrix of stoichiometric amounts
        of each compound j in reaction i.
    :type reaction_matrix: 2D numpy array

    :param compound_names: List of compound names.
    :type compound_names: list of strings

    :returns: List of strings corresponding to each reaction.
    :rtype: list of strings
    """
    reaction_strings = []
    for reaction in reaction_matrix:
        lhs, rhs = ("", "")
        for i, coefficient in enumerate(reaction):
            if coefficient < -1.0e-10:
                if len(lhs) > 0:
                    lhs += " + "
                lhs += f"{-coefficient} {compound_names[i]}"
            if coefficient > 1.0e-10:
                if len(rhs) > 0:
                    rhs += " + "
                rhs += f"{coefficient} {compound_names[i]}"
        reaction_strings.append(f"{lhs} = {rhs}")

    return reaction_strings



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit
# for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import

import importlib
import numpy as np
from fractions import Fraction
from scipy.spatial import Delaunay
from scipy.special import comb
from copy import copy

from .material import cached_property

from ..utils.math import independent_row_indices


# Try to import pycddlib.
# First, try separating the imports into float and
# fractional, then fall back to only using the float
# representation, and finally don't import anything
# (limiting functionality)
try:
    cdd_float = importlib.import_module("cdd")
    cdd_fraction = importlib.import_module("cdd.gmp")
    cdd_gmp_loaded = True
except ImportError:
    try:
        cdd_float = importlib.import_module("cdd")
        cdd_fraction = importlib.import_module("cdd")
        cdd_gmp_loaded = False
    except ImportError:
        cdd_float = None


class SimplexGrid(object):
    """
    A class that creates objects that can efficiently generate a set of points
    that grid a simplex with a user-defined number of vertices. The class
    contains both a generator method and a grid method. It also contains
    an n_points attribute that returns the number of points in the gridded
    simplex.

    This class is available as :class:`burnman.polytope.SimplexGrid`.
    """

    def __init__(self, vertices, points_per_edge):
        """
        Initialize SimplexGrid object with the desired number of vertices
        and points per edge.
        """
        assert vertices >= 2, "need at least two vertices"
        assert points_per_edge >= 2, "need at least 2 points per edge"

        self.vertices = vertices
        self.points_per_edge = points_per_edge

    def generate(self, generate_type="list"):
        """
        Generates the grid points of the simplex in lexicographic order.

        :param generate_type: Determines whether the generator returns
            lists or arrays corresponding to each point in the simplex grid.
            Valid options are 'list' or 'array'.
        :type generate_type: str

        :returns: Grid points of the simplex.
        :rtype: generator of lists or ndarrays (int, ndim=1)
        """

        if generate_type == "list":
            x = [0] * self.vertices
        elif generate_type == "array":
            x = np.zeros(self.vertices, dtype=int)
        else:
            raise Exception("generate_type must be of type list or array.")

        x[self.vertices - 1] = self.points_per_edge - 1

        h = self.vertices
        while True:
            yield copy(x)

            h -= 1
            if h == 0:
                return

            val = x[h]
            x[h] = 0
            x[self.vertices - 1] = val - 1
            x[h - 1] += 1
            if val != 1:
                h = self.vertices

    def grid(self, generate_type="list"):
        """
        Returns either a list or a numpy array
        corresponding the the points in the simplex grid, depending on
        whether the user chooses 'list' (default) or 'array' as
        the generate_type parameter.
        """
        if generate_type == "list":
            return list(self.generate(generate_type))
        elif generate_type == "array":
            return np.array(list(self.generate(generate_type)))
        else:
            raise Exception("generate_type must be of type list or array.")

    def n_points(self):
        """
        The number of points corresponding to the number of vertices and
        points per edge chosen by the user.
        """
        return comb(
            self.vertices + self.points_per_edge - 2, self.vertices - 1, exact=True
        )


class MaterialPolytope(object):
    """
    A class for creating and manipulating polytope objects using pycddlib.

    This class is available as :class:`burnman.polytope.MaterialPolytope`.
    """

    def __init__(
        self,
        equalities,
        inequalities,
        return_fractions=False,
        independent_endmember_occupancies=None,
    ):
        """
        Initialization function for the MaterialPolytope class.
        Declares basis attributes of the class.

        :param equalities: A numpy array containing all the
            equalities defining the polytope. Each row should evaluate to 0.
        :type equalities: numpy.array (2D) of floats or Fractions
        :param inequalities: A numpy array containing all the inequalities
            defining the polytope. Each row should evaluate to <= 0.
        :type inequalities: numpy.array (2D) of the same type as equalities
        :param return_fractions: Whether or not to return fractions.
        :type return_fractions: bool
        :param independent_endmember_occupancies: If specified, this array
            provides the independent endmember set against which the
            dependent endmembers are defined.
        :type independent_endmember_occupancies: numpy.array (2D) or None
        """
        if cdd_float is None:
            raise ImportError(
                "You need to install pycddlib to create a MaterialPolytope object."
            )

        if equalities.dtype != inequalities.dtype:
            raise Exception(
                f"The equalities and inequalities arrays should have the same type ({equalities.dtype} != {inequalities.dtype})."
            )

        self.set_return_type(return_fractions)
        self.equality_matrix = equalities[:, 1:]
        self.equality_vector = -equalities[:, 0]

        if equalities.dtype == Fraction and cdd_gmp_loaded is True:
            self.number_type = Fraction
            self.cdd = cdd_fraction
        elif equalities.dtype == float or cdd_gmp_loaded is False:
            self.number_type = float
            self.cdd = cdd_float
        else:
            raise Exception(
                "equalities should be arrays of either floats or Fractions."
            )

        self.polytope_matrix = self.cdd.matrix_from_array(equalities)
        self.polytope_matrix.lin_set = set(range(len(equalities)))
        self.polytope_matrix.rep_type = self.cdd.RepType.INEQUALITY

        self.cdd.matrix_append_to(
            self.polytope_matrix, self.cdd.matrix_from_array(inequalities)
        )
        self.polytope = self.cdd.polyhedron_from_matrix(self.polytope_matrix)

        if independent_endmember_occupancies is not None:
            self.independent_endmember_occupancies = independent_endmember_occupancies

    def set_return_type(self, return_fractions=False):
        """
        Sets the return_type for the polytope object. Also deletes the cached
        endmember_occupancies property.

        :param return_fractions: Choose whether the generated polytope object
            should return fractions or floats.
        :type return_fractions: bool
        """
        try:
            del self.__dict__["endmember_occupancies"]
        except KeyError:
            pass
        self.return_fractions = return_fractions

    @cached_property
    def raw_vertices(self):
        """
        Returns a list of the vertices of the polytope without any
        postprocessing. See also endmember_occupancies.
        """
        return self.cdd.copy_generators(self.polytope).array

    @cached_property
    def limits(self):
        """
        Return the limits of the polytope (the set of bounding inequalities).
        """
        return np.array(self.cdd.copy_inequalities(self.polytope).array, dtype=float)

    @cached_property
    def n_endmembers(self):
        """
        Return the number of endmembers
        (the number of vertices of the polytope).
        """
        return len(self.raw_vertices)

    @cached_property
    def endmember_occupancies(self):
        """
        Return the endmember occupancies
        (a processed list of all of the vertex locations).
        """
        if self.number_type == float and self.return_fractions:
            v = np.array(
                [
                    [Fraction(value).limit_denominator(1000000) for value in v]
                    for v in self.raw_vertices
                ]
            )
        elif self.number_type == Fraction and not self.return_fractions:
            v = np.array(self.raw_vertices).astype(float)
        else:
            v = np.array(self.raw_vertices)

        if len(v.shape) == 1:
            raise ValueError(
                "The combined equality and positivity "
                "constraints result in a null polytope."
            )
        return v[:, 1:] / v[:, 0, np.newaxis]

    @cached_property
    def independent_endmember_occupancies(self):
        """
        Return an independent set of endmember occupancies
        (a linearly-independent set of vertex locations)
        """
        arr = self.endmember_occupancies
        return arr[independent_row_indices(arr)]

    @cached_property
    def endmembers_as_independent_endmember_amounts(self):
        """
        Return a list of all the endmembers as a linear sum of
        the independent endmembers.
        """
        ind = self.independent_endmember_occupancies

        sol = (
            np.linalg.lstsq(
                np.array(ind.T).astype(float),
                np.array(self.endmember_occupancies.T).astype(float),
                rcond=0,
            )[0]
            .round(decimals=12)
            .T
        )
        return sol

    def _decompose_vertices_into_simplices(self, vertices):
        """
        Decomposes a set of vertices into simplices by Delaunay triangulation.
        """
        # Delaunay triangulation only works in dimensions > 1
        # and we remove the nullspace (sum(fractions) = 1)
        if len(vertices) > 2:
            nulls = np.repeat(vertices[:, -1], vertices.shape[1]).reshape(
                vertices.shape
            )
            tri = Delaunay((vertices - nulls)[:, :-1])
            return tri.simplices
        else:
            return [[0, 1]]

    @cached_property
    def independent_endmember_polytope(self):
        """
        Returns the polytope expressed in terms of proportions of the
        independent endmembers. The polytope involves the first
        n-1 independent endmembers. The last endmember proportion makes
        the sum equal to one.
        """
        arr = self.endmembers_as_independent_endmember_amounts
        arr = np.hstack((np.ones((len(arr), 1)), arr[:, :-1]))
        arr = [[Fraction(value).limit_denominator(1000000) for value in v] for v in arr]
        M = cdd_fraction.matrix_from_array(arr)
        M.rep_type = cdd_fraction.RepType.GENERATOR
        return cdd_fraction.polyhedron_from_matrix(M)

    @cached_property
    def independent_endmember_limits(self):
        """
        Gets the limits of the polytope as a function of the independent
        endmembers.
        """
        ind_poly = self.independent_endmember_polytope
        inequalities = cdd_fraction.copy_inequalities(ind_poly).array
        return np.array(inequalities, dtype=float)

    def subpolytope_from_independent_endmember_limits(self, limits):
        """
        Returns a smaller polytope by applying additional limits to the amounts
        of the independent endmembers.
        """
        ind_poly = self.independent_endmember_polytope
        modified_limits = cdd_fraction.copy_inequalities(ind_poly)
        cdd_fraction.matrix_append_to(
            modified_limits, cdd_fraction.matrix_from_array(limits)
        )
        return cdd_fraction.polyhedron_from_matrix(modified_limits)

    def subpolytope_from_site_occupancy_limits(self, limits):
        """
        Returns a smaller polytope by applying additional limits to the
        individual site occupancies.
        """
        modified_limits = copy(self.polytope_matrix)
        self.cdd.matrix_append_to(modified_limits, self.cdd.matrix_from_array(limits))
        return self.cdd.polyhedron_from_matrix(modified_limits)

    def grid(
        self,
        points_per_edge=2,
        unique_sorted=True,
        grid_type="independent endmember proportions",
        limits=None,
    ):
        """
        Create a grid of points which span the polytope.

        :param points_per_edge: Number of points per edge of the polytope.
        :type points_per_edge: int
        :param unique_sorted: The gridding is done by splitting the polytope
            into a set of simplices. This means that points will be duplicated
            along vertices, faces etc. If unique_sorted is True, this function
            will sort and make the points unique. This is an expensive
            operation for large polytopes, and may not always be necessary.
        :type unique_sorted: bool
        :param grid_type: Whether to grid the polytope in terms of
            independent endmember proportions or site occupancies.
            Choices are 'independent endmember proportions' or
            'site occupancies'
        :type grid_type: str
        :param limits: Additional inequalities restricting the
            gridded area of the polytope.
        :type limits: numpy.array (2D)

        :returns: A list of points gridding the polytope.
        :rtype: numpy.array (2D)
        """
        if limits is None:
            if grid_type == "independent endmember proportions":
                f_occ = self.endmembers_as_independent_endmember_amounts / (
                    points_per_edge - 1
                )
            elif grid_type == "site occupancies":
                f_occ = self.endmember_occupancies / (points_per_edge - 1)
            else:
                raise Exception(
                    "grid type not recognised. Should be one of "
                    "independent endmember proportions "
                    "or site occupancies"
                )

            simplices = self._decompose_vertices_into_simplices(
                self.endmembers_as_independent_endmember_amounts
            )
        else:
            if grid_type == "independent endmember proportions":
                plims = self.subpolytope_from_site_occupancy_limits(limits)
                ppns = np.array(self.cdd.copy_generators(plims).array)[:, 1:]
                last_ppn = np.array([1.0 - sum(p) for p in ppns]).reshape(
                    (len(ppns), 1)
                )
                vertices_as_independent_endmember_proportions = np.hstack(
                    (ppns, last_ppn)
                )
                f_occ = vertices_as_independent_endmember_proportions / (
                    points_per_edge - 1
                )

            elif grid_type == "site occupancies":
                plims = self.subpolytope_from_site_occupancy_limits(limits)
                occ = np.array(self.cdd.copy_generators(plims).array)[:, 1:]
                f_occ = occ / (points_per_edge - 1)

                ind = self.independent_endmember_occupancies

                vertices_as_independent_endmember_proportions = (
                    np.linalg.lstsq(
                        np.array(ind.T).astype(float),
                        np.array(occ.T).astype(float),
                        rcond=None,
                    )[0]
                    .round(decimals=12)
                    .T
                )
            else:
                raise Exception(
                    "grid_type not recognised. "
                    "Should be one of "
                    "independent endmember proportions "
                    "or site occupancies"
                )

            simplices = self._decompose_vertices_into_simplices(
                vertices_as_independent_endmember_proportions
            )

        n_ind = f_occ.shape[1]
        n_simplices = len(simplices)
        dim = len(simplices[0])
        simplex_grid = SimplexGrid(dim, points_per_edge)
        grid = simplex_grid.grid("array")
        points_per_simplex = simplex_grid.n_points()
        n_points = n_simplices * points_per_simplex

        points = np.empty((n_points, n_ind))
        idx = 0
        for i in range(0, n_simplices):
            points[idx : idx + points_per_simplex] = grid.dot(f_occ[simplices[i]])
            idx += points_per_simplex

        if unique_sorted:
            points = np.unique(points, axis=0)
        return points



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit
# for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


"""
SLB_2011
^^^^^^^^

Minerals from Stixrude & Lithgow-Bertelloni 2011 and references therein.
File autogenerated using SLBdata_to_burnman.py.
"""

from __future__ import absolute_import

from ..classes.mineral import Mineral
from ..classes.solution import Solution
from ..classes.solutionmodel import (
    IdealSolution,
    SymmetricRegularSolution,
    AsymmetricRegularSolution,
)
from ..utils.chemistry import dictionarize_formula, formula_mass

"""
SOLID SOLUTIONS
from inv251010 of HeFESTo
"""


class c2c_pyroxene(Solution):
    def __init__(self, molar_fractions=None):
        self.name = "C2/c pyroxene"
        self.solution_model = IdealSolution(
            endmembers=[
                [hp_clinoenstatite(), "[Mg]2Si2O6"],
                [hp_clinoferrosilite(), "[Fe]2Si2O6"],
            ]
        )

        Solution.__init__(self, molar_fractions=molar_fractions)


class ca_ferrite_structured_phase(Solution):
    def __init__(self, molar_fractions=None):
        self.name = "calcium ferrite structured phase"
        self.solution_model = IdealSolution(
            endmembers=[
                [mg_ca_ferrite(), "[Mg]Al[Al]O4"],
                [fe_ca_ferrite(), "[Fe]Al[Al]O4"],
                [na_ca_ferrite(), "[Na]Al[Si]O4"],
            ]
        )

        Solution.__init__(self, molar_fractions=molar_fractions)


class clinopyroxene(Solution):
    def __init__(self, molar_fractions=None):
        self.name = "clinopyroxene"
        self.solution_model = AsymmetricRegularSolution(
            endmembers=[
                [diopside(), "[Ca][Mg][Si]2O6"],
                [hedenbergite(), "[Ca][Fe][Si]2O6"],
                [clinoenstatite(), "[Mg][Mg][Si]2O6"],
                [ca_tschermaks(), "[Ca][Al][Si1/2Al1/2]2O6"],
                [jadeite(), "[Na][Al][Si]2O6"],
            ],
            alphas=[1.0, 1.0, 1.0, 3.5, 1.0],
            energy_interaction=[
                [0.0, 24.74e3, 26.0e3, 24.3e3],
                [24.74e3, 0.0, 0.0e3],
                [60.53136e3, 0.0],
                [10.0e3],
            ],
        )

        Solution.__init__(self, molar_fractions=molar_fractions)


class garnet(Solution):
    def __init__(self, molar_fractions=None):
        self.name = "garnet"
        self.solution_model = SymmetricRegularSolution(
            endmembers=[
                [pyrope(), "[Mg]3[Al][Al]Si3O12"],
                [almandine(), "[Fe]3[Al][Al]Si3O12"],
                [grossular(), "[Ca]3[Al][Al]Si3O12"],
                [mg_majorite(), "[Mg]3[Mg][Si]Si3O12"],
                [jd_majorite(), "[Na2/3Al1/3]3[Al][Si]Si3O12"],
            ],
            energy_interaction=[
                [0.0, 30.0e3, 21.20278e3, 0.0],
                [0.0, 0.0, 0.0],
                [57.77596e3, 0.0],
                [0.0],
            ],
        )

        Solution.__init__(self, molar_fractions=molar_fractions)


class akimotoite(Solution):
    def __init__(self, molar_fractions=None):
        self.name = "akimotoite/ilmenite"
        self.solution_model = SymmetricRegularSolution(
            endmembers=[
                [mg_akimotoite(), "[Mg][Si]O3"],
                [fe_akimotoite(), "[Fe][Si]O3"],
                [corundum(), "[Al][Al]O3"],
            ],
            energy_interaction=[[0.0, 66.0e3], [0.0e3]],
        )

        Solution.__init__(self, molar_fractions=molar_fractions)


class ferropericlase(Solution):
    def __init__(self, molar_fractions=None):
        self.name = "magnesiowustite/ferropericlase"
        self.solution_model = SymmetricRegularSolution(
            endmembers=[[periclase(), "[Mg]O"], [wuestite(), "[Fe]O"]],
            energy_interaction=[[13.0e3]],
        )

        Solution.__init__(self, molar_fractions=molar_fractions)


class mg_fe_olivine(Solution):
    def __init__(self, molar_fractions=None):
        self.name = "olivine"
        self.solution_model = SymmetricRegularSolution(
            endmembers=[[forsterite(), "[Mg]2SiO4"], [fayalite(), "[Fe]2SiO4"]],
            energy_interaction=[[7.81322e3]],
        )

        Solution.__init__(self, molar_fractions=molar_fractions)


class orthopyroxene(Solution):
    def __init__(self, molar_fractions=None):
        self.name = "orthopyroxene"
        self.solution_model = SymmetricRegularSolution(
            endmembers=[
                [enstatite(), "[Mg][Mg]SiSiO6"],
                [ferrosilite(), "[Fe][Fe]SiSiO6"],
                [mg_tschermaks(), "[Mg][Al]AlSiO6"],
                [ortho_diopside(), "[Ca][Mg]SiSiO6"],
            ],  # Al avoidance, see Figure 3
            energy_interaction=[[0.0, 0.0, 32.11352e3], [0.0, 0.0], [48.35316e3]],
        )

        Solution.__init__(self, molar_fractions=molar_fractions)


class plagioclase(Solution):
    def __init__(self, molar_fractions=None):
        self.name = "plagioclase"
        self.solution_model = SymmetricRegularSolution(
            endmembers=[
                [anorthite(), "[Ca][Al]2Si2O8"],
                [albite(), "[Na][Al1/2Si1/2]2Si2O8"],
            ],
            energy_interaction=[[26.0e3]],
        )

        Solution.__init__(self, molar_fractions=molar_fractions)


class post_perovskite(Solution):
    def __init__(self, molar_fractions=None):
        self.name = "post-perovskite/bridgmanite"
        self.solution_model = SymmetricRegularSolution(
            endmembers=[
                [mg_post_perovskite(), "[Mg][Si]O3"],
                [fe_post_perovskite(), "[Fe][Si]O3"],
                [al_post_perovskite(), "[Al][Al]O3"],
            ],
            energy_interaction=[[0.0, 60.0e3], [0.0]],
        )

        Solution.__init__(self, molar_fractions=molar_fractions)


class mg_fe_perovskite(Solution):
    def __init__(self, molar_fractions=None):
        self.name = "magnesium silicate perovskite/bridgmanite"
        self.solution_model = AsymmetricRegularSolution(
            endmembers=[
                [mg_perovskite(), "[Mg][Si]O3"],
                [fe_perovskite(), "[Fe][Si]O3"],
                [al_perovskite(), "[Al][Al]O3"],
            ],
            energy_interaction=[[0.0, 116.0e3], [0.0]],
            alphas=[1.0, 1.0, 0.39],
        )
        Solution.__init__(self, molar_fractions=molar_fractions)


class mg_fe_ringwoodite(Solution):
    def __init__(self, molar_fractions=None):
        self.name = "ringwoodite"
        self.solution_model = SymmetricRegularSolution(
            endmembers=[
                [mg_ringwoodite(), "[Mg]2SiO4"],
                [fe_ringwoodite(), "[Fe]2SiO4"],
            ],
            energy_interaction=[[9.34084e3]],
        )

        Solution.__init__(self, molar_fractions=molar_fractions)


class mg_fe_aluminous_spinel(Solution):
    def __init__(self, molar_fractions=None):
        self.name = "spinel-hercynite binary, fixed order"
        self.solution_model = SymmetricRegularSolution(
            endmembers=[
                [spinel(), "[Mg3/4Al1/4]4[Al7/8Mg1/8]8O16"],
                [hercynite(), "[Fe3/4Al1/4]4[Al7/8Fe1/8]8O16"],
            ],
            energy_interaction=[[5.87646e3]],
        )

        Solution.__init__(self, molar_fractions=molar_fractions)


class mg_fe_wadsleyite(Solution):
    def __init__(self, molar_fractions=None):
        self.name = "wadsleyite"
        self.solution_model = SymmetricRegularSolution(
            endmembers=[
                [mg_wadsleyite(), "[Mg]2SiO4"],
                [fe_wadsleyite(), "[Fe]2SiO4"],
            ],
            energy_interaction=[[16.74718e3]],
        )

        Solution.__init__(self, molar_fractions=molar_fractions)


"""
ENDMEMBERS
"""


class anorthite(Mineral):
    def __init__(self):
        formula = "CaAl2Si2O8"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Anorthite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -4014619.0,
            "V_0": 0.00010061,
            "K_0": 84089150000.0,
            "Kprime_0": 4.0,
            "Debye_0": 752.3911,
            "grueneisen_0": 0.39241,
            "q_0": 1.0,
            "G_0": 39900000000.0,
            "Gprime_0": 1.09134,
            "eta_s_0": 1.6254,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {
            "err_F_0": 4000.0,
            "err_V_0": 0.0,
            "err_K_0": 5000000000.0,
            "err_K_prime_0": 1.0,
            "err_Debye_0": 2.0,
            "err_grueneisen_0": 0.05,
            "err_q_0": 1.0,
            "err_G_0": 3000000000.0,
            "err_Gprime_0": 0.5,
            "err_eta_s_0": 1.0,
        }
        Mineral.__init__(self)


class albite(Mineral):
    def __init__(self):
        formula = "NaAlSi3O8"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Albite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -3718799.0,
            "V_0": 0.000100452,
            "K_0": 59761620000.0,
            "Kprime_0": 4.0,
            "Debye_0": 713.7824,
            "grueneisen_0": 0.56704,
            "q_0": 1.0,
            "G_0": 36000000000.0,
            "Gprime_0": 1.3855,
            "eta_s_0": 1.04208,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {
            "err_F_0": 5000.0,
            "err_V_0": 0.0,
            "err_K_0": 5000000000.0,
            "err_K_prime_0": 1.0,
            "err_Debye_0": 13.0,
            "err_grueneisen_0": 0.03,
            "err_q_0": 1.0,
            "err_G_0": 5000000000.0,
            "err_Gprime_0": 0.5,
            "err_eta_s_0": 1.0,
        }
        Mineral.__init__(self)


class spinel(Mineral):
    def __init__(self):
        formula = "Mg4Al8O16"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Spinel",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -8667568.0,
            "V_0": 0.000159048,
            "K_0": 1.969428e11,
            "Kprime_0": 5.68282,
            "Debye_0": 842.8104,
            "grueneisen_0": 1.02283,
            "q_0": 2.71208,
            "G_0": 1.085e11,
            "Gprime_0": 0.37303,
            "eta_s_0": 2.66282,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.property_modifiers = [
            ["linear", {"delta_E": 0.0, "delta_S": 43.76, "delta_V": 0.0}]
        ]

        self.uncertainties = {
            "err_F_0": 32000.0,
            "err_V_0": 0.0,
            "err_K_0": 1000000000.0,
            "err_K_prime_0": 0.2,
            "err_Debye_0": 33.0,
            "err_grueneisen_0": 0.04,
            "err_q_0": 0.6,
            "err_G_0": 10000000000.0,
            "err_Gprime_0": 0.5,
            "err_eta_s_0": 0.6,
        }
        Mineral.__init__(self)


class hercynite(Mineral):
    def __init__(self):
        formula = "Fe4Al8O16"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Hercynite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -7324009.0,
            "V_0": 0.000163372,
            "K_0": 2.088965e11,
            "Kprime_0": 5.68282,
            "Debye_0": 763.231,
            "grueneisen_0": 1.21719,
            "q_0": 2.71208,
            "G_0": 84500000000.0,
            "Gprime_0": 0.37303,
            "eta_s_0": 2.768,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.property_modifiers = [
            ["linear", {"delta_E": 0.0, "delta_S": 97.28, "delta_V": 0.0}]
        ]

        self.uncertainties = {
            "err_F_0": 35000.0,
            "err_V_0": 0.0,
            "err_K_0": 2000000000.0,
            "err_K_prime_0": 1.0,
            "err_Debye_0": 32.0,
            "err_grueneisen_0": 0.07,
            "err_q_0": 1.0,
            "err_G_0": 13000000000.0,
            "err_Gprime_0": 0.5,
            "err_eta_s_0": 1.0,
        }
        Mineral.__init__(self)


class forsterite(Mineral):
    def __init__(self):
        formula = "Mg2SiO4"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Forsterite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -2055403.0,
            "V_0": 4.3603e-05,
            "K_0": 1.279555e11,
            "Kprime_0": 4.21796,
            "Debye_0": 809.1703,
            "grueneisen_0": 0.99282,
            "q_0": 2.10672,
            "G_0": 81599990000.0,
            "Gprime_0": 1.46257,
            "eta_s_0": 2.29972,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {
            "err_F_0": 2000.0,
            "err_V_0": 0.0,
            "err_K_0": 2000000000.0,
            "err_K_prime_0": 0.2,
            "err_Debye_0": 1.0,
            "err_grueneisen_0": 0.03,
            "err_q_0": 0.2,
            "err_G_0": 2000000000.0,
            "err_Gprime_0": 0.1,
            "err_eta_s_0": 0.1,
        }
        Mineral.__init__(self)


class fayalite(Mineral):
    def __init__(self):
        formula = "Fe2SiO4"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Fayalite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -1370519.0,
            "V_0": 4.629e-05,
            "K_0": 1.349622e11,
            "Kprime_0": 4.21796,
            "Debye_0": 618.7007,
            "grueneisen_0": 1.06023,
            "q_0": 3.6466,
            "G_0": 50899990000.0,
            "Gprime_0": 1.46257,
            "eta_s_0": 1.02497,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.property_modifiers = [
            ["linear", {"delta_E": 0.0, "delta_S": 26.76, "delta_V": 0.0}]
        ]

        self.uncertainties = {
            "err_F_0": 1000.0,
            "err_V_0": 0.0,
            "err_K_0": 2000000000.0,
            "err_K_prime_0": 1.0,
            "err_Debye_0": 2.0,
            "err_grueneisen_0": 0.07,
            "err_q_0": 1.0,
            "err_G_0": 2000000000.0,
            "err_Gprime_0": 0.5,
            "err_eta_s_0": 0.6,
        }
        Mineral.__init__(self)


class mg_wadsleyite(Mineral):
    def __init__(self):
        formula = "Mg2SiO4"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Mg_Wadsleyite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -2027837.0,
            "V_0": 4.0515e-05,
            "K_0": 1.686948e11,
            "Kprime_0": 4.3229,
            "Debye_0": 843.4973,
            "grueneisen_0": 1.2061,
            "q_0": 2.0188,
            "G_0": 1.12e11,
            "Gprime_0": 1.44424,
            "eta_s_0": 2.63683,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {
            "err_F_0": 2000.0,
            "err_V_0": 0.0,
            "err_K_0": 3000000000.0,
            "err_K_prime_0": 0.2,
            "err_Debye_0": 7.0,
            "err_grueneisen_0": 0.09,
            "err_q_0": 1.0,
            "err_G_0": 2000000000.0,
            "err_Gprime_0": 0.2,
            "err_eta_s_0": 0.4,
        }
        Mineral.__init__(self)


class fe_wadsleyite(Mineral):
    def __init__(self):
        formula = "Fe2SiO4"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Fe_Wadsleyite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -1364668.0,
            "V_0": 4.28e-05,
            "K_0": 1.68591e11,
            "Kprime_0": 4.3229,
            "Debye_0": 665.4492,
            "grueneisen_0": 1.2061,
            "q_0": 2.0188,
            "G_0": 72000000000.0,
            "Gprime_0": 1.44424,
            "eta_s_0": 1.04017,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.property_modifiers = [
            ["linear", {"delta_E": 0.0, "delta_S": 26.76, "delta_V": 0.0}]
        ]

        self.uncertainties = {
            "err_F_0": 7000.0,
            "err_V_0": 0.0,
            "err_K_0": 13000000000.0,
            "err_K_prime_0": 1.0,
            "err_Debye_0": 21.0,
            "err_grueneisen_0": 0.3,
            "err_q_0": 1.0,
            "err_G_0": 12000000000.0,
            "err_Gprime_0": 0.5,
            "err_eta_s_0": 1.0,
        }
        Mineral.__init__(self)


class mg_ringwoodite(Mineral):
    def __init__(self):
        formula = "Mg2SiO4"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Mg_Ringwoodite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -2017557.0,
            "V_0": 3.9493e-05,
            "K_0": 1.849009e11,
            "Kprime_0": 4.22035,
            "Debye_0": 877.7094,
            "grueneisen_0": 1.10791,
            "q_0": 2.3914,
            "G_0": 1.23e11,
            "Gprime_0": 1.35412,
            "eta_s_0": 2.30461,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {
            "err_F_0": 2000.0,
            "err_V_0": 0.0,
            "err_K_0": 2000000000.0,
            "err_K_prime_0": 0.2,
            "err_Debye_0": 8.0,
            "err_grueneisen_0": 0.1,
            "err_q_0": 0.4,
            "err_G_0": 2000000000.0,
            "err_Gprime_0": 0.1,
            "err_eta_s_0": 0.5,
        }
        Mineral.__init__(self)


class fe_ringwoodite(Mineral):
    def __init__(self):
        formula = "Fe2SiO4"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Fe_Ringwoodite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -1362772.0,
            "V_0": 4.186e-05,
            "K_0": 2.13412e11,
            "Kprime_0": 4.22035,
            "Debye_0": 677.7177,
            "grueneisen_0": 1.27193,
            "q_0": 2.3914,
            "G_0": 92000000000.0,
            "Gprime_0": 1.35412,
            "eta_s_0": 1.77249,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.property_modifiers = [
            ["linear", {"delta_E": 0.0, "delta_S": 26.76, "delta_V": 0.0}]
        ]

        self.uncertainties = {
            "err_F_0": 2000.0,
            "err_V_0": 0.0,
            "err_K_0": 7000000000.0,
            "err_K_prime_0": 1.0,
            "err_Debye_0": 8.0,
            "err_grueneisen_0": 0.23,
            "err_q_0": 1.0,
            "err_G_0": 10000000000.0,
            "err_Gprime_0": 0.5,
            "err_eta_s_0": 1.0,
        }
        Mineral.__init__(self)


class enstatite(Mineral):
    def __init__(self):
        formula = "Mg2Si2O6"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Enstatite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -2913596.0,
            "V_0": 6.2676e-05,
            "K_0": 1.070768e11,
            "Kprime_0": 7.02751,
            "Debye_0": 812.1848,
            "grueneisen_0": 0.78479,
            "q_0": 3.43846,
            "G_0": 76800000000.0,
            "Gprime_0": 1.54596,
            "eta_s_0": 2.50453,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {
            "err_F_0": 2000.0,
            "err_V_0": 0.0,
            "err_K_0": 2000000000.0,
            "err_K_prime_0": 0.4,
            "err_Debye_0": 4.0,
            "err_grueneisen_0": 0.04,
            "err_q_0": 0.4,
            "err_G_0": 1000000000.0,
            "err_Gprime_0": 0.1,
            "err_eta_s_0": 0.1,
        }
        Mineral.__init__(self)


class ferrosilite(Mineral):
    def __init__(self):
        formula = "Fe2Si2O6"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Ferrosilite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -2225718.0,
            "V_0": 6.5941e-05,
            "K_0": 1.005386e11,
            "Kprime_0": 7.02751,
            "Debye_0": 674.4769,
            "grueneisen_0": 0.71889,
            "q_0": 3.43846,
            "G_0": 52000000000.0,
            "Gprime_0": 1.54596,
            "eta_s_0": 1.07706,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.property_modifiers = [
            ["linear", {"delta_E": 0.0, "delta_S": 26.76, "delta_V": 0.0}]
        ]

        self.uncertainties = {
            "err_F_0": 4000.0,
            "err_V_0": 0.0,
            "err_K_0": 4000000000.0,
            "err_K_prime_0": 0.5,
            "err_Debye_0": 10.0,
            "err_grueneisen_0": 0.08,
            "err_q_0": 1.0,
            "err_G_0": 5000000000.0,
            "err_Gprime_0": 0.5,
            "err_eta_s_0": 1.0,
        }
        Mineral.__init__(self)


class mg_tschermaks(Mineral):
    def __init__(self):
        formula = "MgAl2SiO6"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Mg_Tschermaks",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -3002470.0,
            "V_0": 5.914e-05,
            "K_0": 1.070768e11,
            "Kprime_0": 7.02751,
            "Debye_0": 783.8404,
            "grueneisen_0": 0.78479,
            "q_0": 3.43846,
            "G_0": 95950860000.0,
            "Gprime_0": 1.54596,
            "eta_s_0": 2.49099,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {
            "err_F_0": 9000.0,
            "err_V_0": 0.0,
            "err_K_0": 10000000000.0,
            "err_K_prime_0": 1.0,
            "err_Debye_0": 24.0,
            "err_grueneisen_0": 0.3,
            "err_q_0": 1.0,
            "err_G_0": 10000000000.0,
            "err_Gprime_0": 0.5,
            "err_eta_s_0": 1.0,
        }
        Mineral.__init__(self)


class ortho_diopside(Mineral):
    def __init__(self):
        formula = "CaMgSi2O6"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Ortho_Diopside",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -3015827.0,
            "V_0": 6.8054e-05,
            "K_0": 1.070768e11,
            "Kprime_0": 7.02751,
            "Debye_0": 744.6988,
            "grueneisen_0": 0.78479,
            "q_0": 3.43846,
            "G_0": 58458950000.0,
            "Gprime_0": 1.54596,
            "eta_s_0": 1.36161,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {
            "err_F_0": 3000.0,
            "err_V_0": 0.0,
            "err_K_0": 10000000000.0,
            "err_K_prime_0": 1.0,
            "err_Debye_0": 9.0,
            "err_grueneisen_0": 0.3,
            "err_q_0": 1.0,
            "err_G_0": 10000000000.0,
            "err_Gprime_0": 0.5,
            "err_eta_s_0": 1.0,
        }
        Mineral.__init__(self)


class diopside(Mineral):
    def __init__(self):
        formula = "CaMgSi2O6"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Diopside",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -3029531.0,
            "V_0": 6.6039e-05,
            "K_0": 1.122413e11,
            "Kprime_0": 5.23885,
            "Debye_0": 781.6146,
            "grueneisen_0": 0.95873,
            "q_0": 1.52852,
            "G_0": 67000000000.0,
            "Gprime_0": 1.37293,
            "eta_s_0": 1.57351,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {
            "err_F_0": 2000.0,
            "err_V_0": 0.0,
            "err_K_0": 5000000000.0,
            "err_K_prime_0": 1.8,
            "err_Debye_0": 3.0,
            "err_grueneisen_0": 0.05,
            "err_q_0": 2.0,
            "err_G_0": 2000000000.0,
            "err_Gprime_0": 0.5,
            "err_eta_s_0": 1.0,
        }
        Mineral.__init__(self)


class hedenbergite(Mineral):
    def __init__(self):
        formula = "CaFeSi2O6"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Hedenbergite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -2677330.0,
            "V_0": 6.7867e-05,
            "K_0": 1.192555e11,
            "Kprime_0": 5.23885,
            "Debye_0": 701.5851,
            "grueneisen_0": 0.93516,
            "q_0": 1.52852,
            "G_0": 61000000000.0,
            "Gprime_0": 1.17647,
            "eta_s_0": 1.5703,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.property_modifiers = [
            ["linear", {"delta_E": 0.0, "delta_S": 13.38, "delta_V": 0.0}]
        ]

        self.uncertainties = {
            "err_F_0": 45000.0,
            "err_V_0": 0.0,
            "err_K_0": 4000000000.0,
            "err_K_prime_0": 1.0,
            "err_Debye_0": 2.0,
            "err_grueneisen_0": 0.06,
            "err_q_0": 1.0,
            "err_G_0": 1000000000.0,
            "err_Gprime_0": 0.5,
            "err_eta_s_0": 1.0,
        }
        Mineral.__init__(self)


class clinoenstatite(Mineral):
    def __init__(self):
        formula = "Mg2Si2O6"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Clinoenstatite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -2905918.0,
            "V_0": 6.25e-05,
            "K_0": 1.122413e11,
            "Kprime_0": 5.23885,
            "Debye_0": 805.0547,
            "grueneisen_0": 0.95873,
            "q_0": 1.52852,
            "G_0": 79496860000.0,
            "Gprime_0": 1.62901,
            "eta_s_0": 1.69074,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {
            "err_F_0": 3000.0,
            "err_V_0": 0.0,
            "err_K_0": 10000000000.0,
            "err_K_prime_0": 1.0,
            "err_Debye_0": 10.0,
            "err_grueneisen_0": 0.3,
            "err_q_0": 1.0,
            "err_G_0": 10000000000.0,
            "err_Gprime_0": 0.5,
            "err_eta_s_0": 1.0,
        }
        Mineral.__init__(self)


class ca_tschermaks(Mineral):
    def __init__(self):
        formula = "CaAl2SiO6"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Ca_Tschermaks",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -3120253.0,
            "V_0": 6.3574e-05,
            "K_0": 1.122413e11,
            "Kprime_0": 5.23885,
            "Debye_0": 803.6626,
            "grueneisen_0": 0.78126,
            "q_0": 1.52852,
            "G_0": 75160660000.0,
            "Gprime_0": 1.54016,
            "eta_s_0": 1.9672,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.property_modifiers = [
            ["linear", {"delta_E": 0.0, "delta_S": 11.525, "delta_V": 0.0}]
        ]

        self.uncertainties = {
            "err_F_0": 5000.0,
            "err_V_0": 0.0,
            "err_K_0": 10000000000.0,
            "err_K_prime_0": 1.0,
            "err_Debye_0": 5.0,
            "err_grueneisen_0": 0.0,
            "err_q_0": 1.0,
            "err_G_0": 10000000000.0,
            "err_Gprime_0": 0.5,
            "err_eta_s_0": 1.0,
        }
        Mineral.__init__(self)


class jadeite(Mineral):
    def __init__(self):
        formula = "NaAlSi2O6"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Jadeite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -2855192.0,
            "V_0": 6.0508e-05,
            "K_0": 1.422873e11,
            "Kprime_0": 5.23885,
            "Debye_0": 820.7623,
            "grueneisen_0": 0.903,
            "q_0": 0.39234,
            "G_0": 85000000000.0,
            "Gprime_0": 1.37398,
            "eta_s_0": 2.18453,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {
            "err_F_0": 3000.0,
            "err_V_0": 0.0,
            "err_K_0": 2000000000.0,
            "err_K_prime_0": 1.0,
            "err_Debye_0": 12.0,
            "err_grueneisen_0": 0.08,
            "err_q_0": 1.4,
            "err_G_0": 2000000000.0,
            "err_Gprime_0": 0.5,
            "err_eta_s_0": 1.0,
        }
        Mineral.__init__(self)


class hp_clinoenstatite(Mineral):
    def __init__(self):
        formula = "Mg2Si2O6"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "HP_Clinoenstatite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -2905788.0,
            "V_0": 6.076e-05,
            "K_0": 1.160254e11,
            "Kprime_0": 6.23685,
            "Debye_0": 824.4439,
            "grueneisen_0": 1.12473,
            "q_0": 0.20401,
            "G_0": 87927170000.0,
            "Gprime_0": 1.84119,
            "eta_s_0": 2.14181,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {
            "err_F_0": 3000.0,
            "err_V_0": 0.0,
            "err_K_0": 1000000000.0,
            "err_K_prime_0": 0.3,
            "err_Debye_0": 7.0,
            "err_grueneisen_0": 0.05,
            "err_q_0": 0.5,
            "err_G_0": 1000000000.0,
            "err_Gprime_0": 0.1,
            "err_eta_s_0": 0.5,
        }
        Mineral.__init__(self)


class hp_clinoferrosilite(Mineral):
    def __init__(self):
        formula = "Fe2Si2O6"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "HP_Clinoferrosilite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -2222183.0,
            "V_0": 6.385413e-05,
            "K_0": 1.160254e11,
            "Kprime_0": 6.23685,
            "Debye_0": 691.564,
            "grueneisen_0": 1.12473,
            "q_0": 0.20401,
            "G_0": 70623090000.0,
            "Gprime_0": 1.84119,
            "eta_s_0": 0.79216,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.property_modifiers = [
            ["linear", {"delta_E": 0.0, "delta_S": 26.76, "delta_V": 0.0}]
        ]

        self.uncertainties = {
            "err_F_0": 4000.0,
            "err_V_0": 0.0,
            "err_K_0": 10000000000.0,
            "err_K_prime_0": 1.0,
            "err_Debye_0": 11.0,
            "err_grueneisen_0": 0.3,
            "err_q_0": 1.0,
            "err_G_0": 10000000000.0,
            "err_Gprime_0": 0.5,
            "err_eta_s_0": 1.0,
        }
        Mineral.__init__(self)


class ca_perovskite(Mineral):
    def __init__(self):
        formula = "CaSiO3"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Ca_Perovskite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -1463358.0,
            "V_0": 2.745e-05,
            "K_0": 2.36e11,
            "Kprime_0": 3.9,
            "Debye_0": 795.779,
            "grueneisen_0": 1.88839,
            "q_0": 0.89769,
            "G_0": 1.568315e11,
            "Gprime_0": 2.22713,
            "eta_s_0": 1.28818,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {
            "err_F_0": 8000.0,
            "err_V_0": 0.0,
            "err_K_0": 4000000000.0,
            "err_K_prime_0": 0.2,
            "err_Debye_0": 44.0,
            "err_grueneisen_0": 0.07,
            "err_q_0": 1.6,
            "err_G_0": 12000000000.0,
            "err_Gprime_0": 0.5,
            "err_eta_s_0": 1.0,
        }
        Mineral.__init__(self)


class mg_akimotoite(Mineral):
    def __init__(self):
        formula = "MgSiO3"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Mg_Akimotoite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -1410850.0,
            "V_0": 2.6354e-05,
            "K_0": 2.10706e11,
            "Kprime_0": 5.62088,
            "Debye_0": 935.9778,
            "grueneisen_0": 1.18984,
            "q_0": 2.34514,
            "G_0": 1.32e11,
            "Gprime_0": 1.57889,
            "eta_s_0": 2.80782,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {
            "err_F_0": 2000.0,
            "err_V_0": 0.0,
            "err_K_0": 4000000000.0,
            "err_K_prime_0": 0.8,
            "err_Debye_0": 12.0,
            "err_grueneisen_0": 0.13,
            "err_q_0": 0.8,
            "err_G_0": 8000000000.0,
            "err_Gprime_0": 0.5,
            "err_eta_s_0": 1.0,
        }
        Mineral.__init__(self)


class fe_akimotoite(Mineral):
    def __init__(self):
        formula = "FeSiO3"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Fe_Akimotoite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -1067598.0,
            "V_0": 2.6854e-05,
            "K_0": 2.10706e11,
            "Kprime_0": 5.62088,
            "Debye_0": 887.8709,
            "grueneisen_0": 1.18984,
            "q_0": 2.34514,
            "G_0": 1.523046e11,
            "Gprime_0": 1.57889,
            "eta_s_0": 3.5716,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.property_modifiers = [
            ["linear", {"delta_E": 0.0, "delta_S": 13.38, "delta_V": 0.0}]
        ]

        self.uncertainties = {
            "err_F_0": 21000.0,
            "err_V_0": 0.0,
            "err_K_0": 10000000000.0,
            "err_K_prime_0": 1.0,
            "err_Debye_0": 120.0,
            "err_grueneisen_0": 0.3,
            "err_q_0": 1.0,
            "err_G_0": 10000000000.0,
            "err_Gprime_0": 0.5,
            "err_eta_s_0": 1.0,
        }
        Mineral.__init__(self)


class corundum(Mineral):
    def __init__(self):
        formula = "AlAlO3"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Corundum",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -1582454.0,
            "V_0": 2.5577e-05,
            "K_0": 2.525457e11,
            "Kprime_0": 4.33728,
            "Debye_0": 932.5696,
            "grueneisen_0": 1.32442,
            "q_0": 1.30316,
            "G_0": 1.632e11,
            "Gprime_0": 1.64174,
            "eta_s_0": 2.8316,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {
            "err_F_0": 1000.0,
            "err_V_0": 0.0,
            "err_K_0": 5000000000.0,
            "err_K_prime_0": 0.2,
            "err_Debye_0": 3.0,
            "err_grueneisen_0": 0.04,
            "err_q_0": 0.2,
            "err_G_0": 2000000000.0,
            "err_Gprime_0": 0.1,
            "err_eta_s_0": 0.2,
        }
        Mineral.__init__(self)


class pyrope(Mineral):
    def __init__(self):
        formula = "Mg3Al2Si3O12"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Pyrope",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -5936538.0,
            "V_0": 0.00011308,
            "K_0": 1.702396e11,
            "Kprime_0": 4.11067,
            "Debye_0": 823.2102,
            "grueneisen_0": 1.01424,
            "q_0": 1.42169,
            "G_0": 93699990000.0,
            "Gprime_0": 1.35756,
            "eta_s_0": 0.98186,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {
            "err_F_0": 10000.0,
            "err_V_0": 0.0,
            "err_K_0": 2000000000.0,
            "err_K_prime_0": 0.3,
            "err_Debye_0": 4.0,
            "err_grueneisen_0": 0.06,
            "err_q_0": 0.5,
            "err_G_0": 2000000000.0,
            "err_Gprime_0": 0.2,
            "err_eta_s_0": 0.3,
        }
        Mineral.__init__(self)


class almandine(Mineral):
    def __init__(self):
        formula = "Fe3Al2Si3O12"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Almandine",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -4935516.0,
            "V_0": 0.00011543,
            "K_0": 1.738963e11,
            "Kprime_0": 4.91341,
            "Debye_0": 741.356,
            "grueneisen_0": 1.06495,
            "q_0": 1.42169,
            "G_0": 96000000000.0,
            "Gprime_0": 1.40927,
            "eta_s_0": 2.09292,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.property_modifiers = [
            ["linear", {"delta_E": 0.0, "delta_S": 40.14, "delta_V": 0.0}]
        ]

        self.uncertainties = {
            "err_F_0": 29000.0,
            "err_V_0": 0.0,
            "err_K_0": 2000000000.0,
            "err_K_prime_0": 0.2,
            "err_Debye_0": 5.0,
            "err_grueneisen_0": 0.06,
            "err_q_0": 1.0,
            "err_G_0": 1000000000.0,
            "err_Gprime_0": 0.1,
            "err_eta_s_0": 1.0,
        }
        Mineral.__init__(self)


class grossular(Mineral):
    def __init__(self):
        formula = "Ca3Al2Si3O12"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Grossular",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -6277935.0,
            "V_0": 0.00012512,
            "K_0": 1.670622e11,
            "Kprime_0": 3.91544,
            "Debye_0": 822.743,
            "grueneisen_0": 1.05404,
            "q_0": 1.88887,
            "G_0": 1.09e11,
            "Gprime_0": 1.16274,
            "eta_s_0": 2.38418,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {
            "err_F_0": 11000.0,
            "err_V_0": 0.0,
            "err_K_0": 1000000000.0,
            "err_K_prime_0": 0.2,
            "err_Debye_0": 2.0,
            "err_grueneisen_0": 0.06,
            "err_q_0": 0.2,
            "err_G_0": 4000000000.0,
            "err_Gprime_0": 0.1,
            "err_eta_s_0": 0.1,
        }
        Mineral.__init__(self)


class mg_majorite(Mineral):
    def __init__(self):
        formula = "Mg4Si4O12"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Mg_Majorite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -5691614.0,
            "V_0": 0.000114324,
            "K_0": 1.651183e11,
            "Kprime_0": 4.21183,
            "Debye_0": 822.458,
            "grueneisen_0": 0.97682,
            "q_0": 1.53581,
            "G_0": 84999990000.0,
            "Gprime_0": 1.42969,
            "eta_s_0": 1.0178,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {
            "err_F_0": 10000.0,
            "err_V_0": 0.0,
            "err_K_0": 3000000000.0,
            "err_K_prime_0": 0.3,
            "err_Debye_0": 4.0,
            "err_grueneisen_0": 0.07,
            "err_q_0": 0.5,
            "err_G_0": 2000000000.0,
            "err_Gprime_0": 0.2,
            "err_eta_s_0": 0.3,
        }
        Mineral.__init__(self)


class jd_majorite(Mineral):
    def __init__(self):
        formula = "Na2Al2Si4O12"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Jd_Majorite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -5518542.0,
            "V_0": 0.00011094,
            "K_0": 1.770772e11,
            "Kprime_0": 4.11067,
            "Debye_0": 895.914,
            "grueneisen_0": 1.01424,
            "q_0": 1.42169,
            "G_0": 1.25e11,
            "Gprime_0": 1.35756,
            "eta_s_0": 3.30517,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {
            "err_F_0": 14000.0,
            "err_V_0": 0.0,
            "err_K_0": 7000000000.0,
            "err_K_prime_0": 1.0,
            "err_Debye_0": 18.0,
            "err_grueneisen_0": 0.3,
            "err_q_0": 1.0,
            "err_G_0": 4000000000.0,
            "err_Gprime_0": 0.5,
            "err_eta_s_0": 1.0,
        }
        Mineral.__init__(self)


class quartz(Mineral):
    def __init__(self):
        formula = "SiO2"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Quartz",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -858853.4,
            "V_0": 2.367003e-05,
            "K_0": 49547430000.0,
            "Kprime_0": 4.33155,
            "Debye_0": 816.3307,
            "grueneisen_0": -0.00296,
            "q_0": 1.0,
            "G_0": 44856170000.0,
            "Gprime_0": 0.95315,
            "eta_s_0": 2.36469,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.property_modifiers = [
            ["landau", {"Tc_0": 847.0, "S_D": 5.164, "V_D": 1.222e-06}]
        ]

        self.uncertainties = {
            "err_F_0": 1000.0,
            "err_V_0": 0.0,
            "err_K_0": 1000000000.0,
            "err_K_prime_0": 0.1,
            "err_Debye_0": 31.0,
            "err_grueneisen_0": 0.05,
            "err_q_0": 1.0,
            "err_G_0": 1000000000.0,
            "err_Gprime_0": 0.1,
            "err_eta_s_0": 1.0,
        }
        Mineral.__init__(self)


class coesite(Mineral):
    def __init__(self):
        formula = "SiO2"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Coesite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -855068.5,
            "V_0": 2.0657e-05,
            "K_0": 1.135856e11,
            "Kprime_0": 4.0,
            "Debye_0": 852.4267,
            "grueneisen_0": 0.39157,
            "q_0": 1.0,
            "G_0": 61600010000.0,
            "Gprime_0": 1.24734,
            "eta_s_0": 2.39793,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {
            "err_F_0": 1000.0,
            "err_V_0": 0.0,
            "err_K_0": 1000000000.0,
            "err_K_prime_0": 1.0,
            "err_Debye_0": 9.0,
            "err_grueneisen_0": 0.05,
            "err_q_0": 1.0,
            "err_G_0": 1000000000.0,
            "err_Gprime_0": 0.5,
            "err_eta_s_0": 1.0,
        }
        Mineral.__init__(self)


class stishovite(Mineral):
    def __init__(self):
        formula = "SiO2"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Stishovite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -818984.6,
            "V_0": 1.4017e-05,
            "K_0": 3.143352e11,
            "Kprime_0": 3.75122,
            "Debye_0": 1107.824,
            "grueneisen_0": 1.37466,
            "q_0": 2.83517,
            "G_0": 2.2e11,
            "Gprime_0": 1.93334,
            "eta_s_0": 4.60904,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.property_modifiers = [
            ["landau", {"Tc_0": -4250.0, "S_D": 0.012, "V_D": 1e-09}]
        ]

        self.uncertainties = {
            "err_F_0": 1000.0,
            "err_V_0": 0.0,
            "err_K_0": 8000000000.0,
            "err_K_prime_0": 0.1,
            "err_Debye_0": 13.0,
            "err_grueneisen_0": 0.17,
            "err_q_0": 2.2,
            "err_G_0": 12000000000.0,
            "err_Gprime_0": 0.1,
            "err_eta_s_0": 1.0,
        }
        Mineral.__init__(self)


class seifertite(Mineral):
    def __init__(self):
        formula = "SiO2"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Seifertite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -794335.4,
            "V_0": 1.367e-05,
            "K_0": 3.275843e11,
            "Kprime_0": 4.01553,
            "Debye_0": 1140.772,
            "grueneisen_0": 1.37466,
            "q_0": 2.83517,
            "G_0": 2.274532e11,
            "Gprime_0": 1.76965,
            "eta_s_0": 4.97108,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {
            "err_F_0": 2000.0,
            "err_V_0": 0.0,
            "err_K_0": 2000000000.0,
            "err_K_prime_0": 0.1,
            "err_Debye_0": 16.0,
            "err_grueneisen_0": 0.3,
            "err_q_0": 1.0,
            "err_G_0": 2000000000.0,
            "err_Gprime_0": 0.1,
            "err_eta_s_0": 1.0,
        }
        Mineral.__init__(self)


class mg_perovskite(Mineral):
    def __init__(self):
        formula = "MgSiO3"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Mg_Perovskite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -1368283.0,
            "V_0": 2.4445e-05,
            "K_0": 2.505264e11,
            "Kprime_0": 4.14,
            "Debye_0": 905.9412,
            "grueneisen_0": 1.56508,
            "q_0": 1.10945,
            "G_0": 1.729e11,
            "Gprime_0": 1.69037,
            "eta_s_0": 2.56536,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {
            "err_F_0": 1000.0,
            "err_V_0": 0.0,
            "err_K_0": 3000000000.0,
            "err_K_prime_0": 0.1,
            "err_Debye_0": 5.0,
            "err_grueneisen_0": 0.05,
            "err_q_0": 0.3,
            "err_G_0": 2000000000.0,
            "err_Gprime_0": 0.0,
            "err_eta_s_0": 0.3,
        }
        Mineral.__init__(self)


class fe_perovskite(Mineral):
    def __init__(self):
        formula = "FeSiO3"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Fe_Perovskite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -1040920.0,
            "V_0": 2.5485e-05,
            "K_0": 2.721152e11,
            "Kprime_0": 4.14,
            "Debye_0": 870.8122,
            "grueneisen_0": 1.56508,
            "q_0": 1.10945,
            "G_0": 1.326849e11,
            "Gprime_0": 1.37485,
            "eta_s_0": 2.29211,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.property_modifiers = [
            ["linear", {"delta_E": 0.0, "delta_S": 13.38, "delta_V": 0.0}]
        ]

        self.uncertainties = {
            "err_F_0": 6000.0,
            "err_V_0": 0.0,
            "err_K_0": 40000000000.0,
            "err_K_prime_0": 1.0,
            "err_Debye_0": 26.0,
            "err_grueneisen_0": 0.3,
            "err_q_0": 1.0,
            "err_G_0": 40000000000.0,
            "err_Gprime_0": 0.0,
            "err_eta_s_0": 1.0,
        }
        Mineral.__init__(self)


class al_perovskite(Mineral):
    def __init__(self):
        formula = "AlAlO3"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Al_perovskite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -1533878.0,
            "V_0": 2.4944e-05,
            "K_0": 2.582e11,
            "Kprime_0": 4.14,
            "Debye_0": 886.4601,
            "grueneisen_0": 1.56508,
            "q_0": 1.10945,
            "G_0": 1.713116e11,
            "Gprime_0": 1.49706,
            "eta_s_0": 2.47126,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {
            "err_F_0": 2000.0,
            "err_V_0": 0.0,
            "err_K_0": 10000000000.0,
            "err_K_prime_0": 0.5,
            "err_Debye_0": 7.0,
            "err_grueneisen_0": 0.3,
            "err_q_0": 1.0,
            "err_G_0": 10000000000.0,
            "err_Gprime_0": 0.1,
            "err_eta_s_0": 0.5,
        }
        Mineral.__init__(self)


class mg_post_perovskite(Mineral):
    def __init__(self):
        formula = "MgSiO3"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Mg_Post_Perovskite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -1348641.0,
            "V_0": 2.4419e-05,
            "K_0": 2.312e11,
            "Kprime_0": 4.0,
            "Debye_0": 855.8173,
            "grueneisen_0": 1.89155,
            "q_0": 1.09081,
            "G_0": 1.50167e11,
            "Gprime_0": 1.97874,
            "eta_s_0": 1.16704,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {
            "err_F_0": 3000.0,
            "err_V_0": 0.0,
            "err_K_0": 1000000000.0,
            "err_K_prime_0": 0.1,
            "err_Debye_0": 7.0,
            "err_grueneisen_0": 0.03,
            "err_q_0": 0.1,
            "err_G_0": 4000000000.0,
            "err_Gprime_0": 0.1,
            "err_eta_s_0": 0.2,
        }
        Mineral.__init__(self)


class fe_post_perovskite(Mineral):
    def __init__(self):
        formula = "FeSiO3"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Fe_Post_Perovskite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -981806.9,
            "V_0": 2.5459e-05,
            "K_0": 2.312e11,
            "Kprime_0": 4.0,
            "Debye_0": 781.3465,
            "grueneisen_0": 1.89155,
            "q_0": 1.09081,
            "G_0": 1.295e11,
            "Gprime_0": 1.44675,
            "eta_s_0": 1.36382,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.property_modifiers = [
            ["linear", {"delta_E": 0.0, "delta_S": 13.38, "delta_V": 0.0}]
        ]

        self.uncertainties = {
            "err_F_0": 21000.0,
            "err_V_0": 0.0,
            "err_K_0": 10000000000.0,
            "err_K_prime_0": 1.0,
            "err_Debye_0": 52.0,
            "err_grueneisen_0": 0.3,
            "err_q_0": 1.0,
            "err_G_0": 5000000000.0,
            "err_Gprime_0": 0.1,
            "err_eta_s_0": 1.0,
        }
        Mineral.__init__(self)


class al_post_perovskite(Mineral):
    def __init__(self):
        formula = "AlAlO3"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Al_Post_Perovskite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -1377582.0,
            "V_0": 2.3847e-05,
            "K_0": 2.49e11,
            "Kprime_0": 4.0,
            "Debye_0": 762.1951,
            "grueneisen_0": 1.64573,
            "q_0": 1.09081,
            "G_0": 91965310000.0,
            "Gprime_0": 1.81603,
            "eta_s_0": 2.83762,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {
            "err_F_0": 4000.0,
            "err_V_0": 0.0,
            "err_K_0": 20000000000.0,
            "err_K_prime_0": 0.1,
            "err_Debye_0": 9.0,
            "err_grueneisen_0": 0.02,
            "err_q_0": 1.0,
            "err_G_0": 10000000000.0,
            "err_Gprime_0": 0.1,
            "err_eta_s_0": 0.2,
        }
        Mineral.__init__(self)


class periclase(Mineral):
    def __init__(self):
        formula = "MgO"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Periclase",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -569444.6,
            "V_0": 1.1244e-05,
            "K_0": 1.613836e11,
            "Kprime_0": 3.84045,
            "Debye_0": 767.0977,
            "grueneisen_0": 1.36127,
            "q_0": 1.7217,
            "G_0": 1.309e11,
            "Gprime_0": 2.1438,
            "eta_s_0": 2.81765,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {
            "err_F_0": 0.0,
            "err_V_0": 0.0,
            "err_K_0": 3000000000.0,
            "err_K_prime_0": 0.2,
            "err_Debye_0": 9.0,
            "err_grueneisen_0": 0.05,
            "err_q_0": 0.2,
            "err_G_0": 1000000000.0,
            "err_Gprime_0": 0.1,
            "err_eta_s_0": 0.2,
        }
        Mineral.__init__(self)


class wuestite(Mineral):
    def __init__(self):
        formula = "FeO"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Wuestite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -242146.0,
            "V_0": 1.2264e-05,
            "K_0": 1.794442e11,
            "Kprime_0": 4.9376,
            "Debye_0": 454.1592,
            "grueneisen_0": 1.53047,
            "q_0": 1.7217,
            "G_0": 59000000000.0,
            "Gprime_0": 1.44673,
            "eta_s_0": -0.05731,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.property_modifiers = [
            ["linear", {"delta_E": 0.0, "delta_S": 13.38, "delta_V": 0.0}]
        ]

        self.uncertainties = {
            "err_F_0": 1000.0,
            "err_V_0": 0.0,
            "err_K_0": 1000000000.0,
            "err_K_prime_0": 0.2,
            "err_Debye_0": 21.0,
            "err_grueneisen_0": 0.13,
            "err_q_0": 1.0,
            "err_G_0": 1000000000.0,
            "err_Gprime_0": 0.1,
            "err_eta_s_0": 1.0,
        }
        Mineral.__init__(self)


class mg_ca_ferrite(Mineral):
    def __init__(self):
        formula = "MgAl2O4"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Mg_Ca_Ferrite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -2122169.0,
            "V_0": 3.6177e-05,
            "K_0": 2.106663e11,
            "Kprime_0": 4.0528,
            "Debye_0": 838.6291,
            "grueneisen_0": 1.31156,
            "q_0": 1.0,
            "G_0": 1.29826e11,
            "Gprime_0": 1.75878,
            "eta_s_0": 2.1073,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {
            "err_F_0": 4000.0,
            "err_V_0": 0.0,
            "err_K_0": 1000000000.0,
            "err_K_prime_0": 0.1,
            "err_Debye_0": 16.0,
            "err_grueneisen_0": 0.3,
            "err_q_0": 1.0,
            "err_G_0": 1000000000.0,
            "err_Gprime_0": 0.1,
            "err_eta_s_0": 1.0,
        }
        Mineral.__init__(self)


class fe_ca_ferrite(Mineral):
    def __init__(self):
        formula = "FeAl2O4"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Fe_Ca_Ferrite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -1790284.0,
            "V_0": 3.7258e-05,
            "K_0": 2.106663e11,
            "Kprime_0": 4.0528,
            "Debye_0": 804.1986,
            "grueneisen_0": 1.31156,
            "q_0": 1.0,
            "G_0": 1.535236e11,
            "Gprime_0": 1.75878,
            "eta_s_0": 3.0268,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.property_modifiers = [
            ["linear", {"delta_E": 0.0, "delta_S": 13.38, "delta_V": 0.0}]
        ]

        self.uncertainties = {
            "err_F_0": 25000.0,
            "err_V_0": 0.0,
            "err_K_0": 10000000000.0,
            "err_K_prime_0": 1.0,
            "err_Debye_0": 69.0,
            "err_grueneisen_0": 0.3,
            "err_q_0": 1.0,
            "err_G_0": 10000000000.0,
            "err_Gprime_0": 0.5,
            "err_eta_s_0": 1.0,
        }
        Mineral.__init__(self)


class na_ca_ferrite(Mineral):
    def __init__(self):
        formula = "NaAlSiO4"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Na_Ca_Ferrite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -1844129.0,
            "V_0": 3.627e-05,
            "K_0": 1.613385e11,
            "Kprime_0": 4.32479,
            "Debye_0": 812.4769,
            "grueneisen_0": 0.69428,
            "q_0": 1.0,
            "G_0": 1.220049e11,
            "Gprime_0": 2.07687,
            "eta_s_0": 2.79016,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {
            "err_F_0": 11000.0,
            "err_V_0": 0.0,
            "err_K_0": 1000000000.0,
            "err_K_prime_0": 0.1,
            "err_Debye_0": 51.0,
            "err_grueneisen_0": 0.3,
            "err_q_0": 1.0,
            "err_G_0": 1000000000.0,
            "err_Gprime_0": 0.1,
            "err_eta_s_0": 1.0,
        }
        Mineral.__init__(self)


class kyanite(Mineral):
    def __init__(self):
        formula = "Al2SiO5"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Kyanite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -2446058.0,
            "V_0": 4.4227e-05,
            "K_0": 1.6e11,
            "Kprime_0": 4.0,
            "Debye_0": 943.1665,
            "grueneisen_0": 0.9255,
            "q_0": 1.0,
            "G_0": 1.204033e11,
            "Gprime_0": 1.7308,
            "eta_s_0": 2.96665,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {
            "err_F_0": 4000.0,
            "err_V_0": 0.0,
            "err_K_0": 1000000000.0,
            "err_K_prime_0": 0.0,
            "err_Debye_0": 8.0,
            "err_grueneisen_0": 0.07,
            "err_q_0": 1.0,
            "err_G_0": 10000000000.0,
            "err_Gprime_0": 0.5,
            "err_eta_s_0": 1.0,
        }
        Mineral.__init__(self)


class nepheline(Mineral):
    def __init__(self):
        formula = "NaAlSiO4"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Nepheline",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -1992104.0,
            "V_0": 5.46684e-05,
            "K_0": 53077990000.0,
            "Kprime_0": 4.0,
            "Debye_0": 700.9422,
            "grueneisen_0": 0.69428,
            "q_0": 1.0,
            "G_0": 30700000000.0,
            "Gprime_0": 1.33031,
            "eta_s_0": 0.6291,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {
            "err_F_0": 3000.0,
            "err_V_0": 0.0,
            "err_K_0": 1000000000.0,
            "err_K_prime_0": 1.0,
            "err_Debye_0": 13.0,
            "err_grueneisen_0": 0.03,
            "err_q_0": 1.0,
            "err_G_0": 1000000000.0,
            "err_Gprime_0": 0.5,
            "err_eta_s_0": 1.0,
        }
        Mineral.__init__(self)


"""
Mineral aliases
"""

# Feldspars
ab = albite
an = anorthite

# LP Spinels
sp = spinel
hc = hercynite

# Olivine polymorphs
fo = forsterite
fa = fayalite
mgwa = mg_wadsleyite
fewa = fe_wadsleyite
mgri = mg_ringwoodite
feri = fe_ringwoodite

# Orthopyroxenes
en = enstatite
fs = ferrosilite
mgts = mg_tschermaks
odi = ortho_diopside

# Clinopyroxenes
di = diopside
he = hedenbergite
cen = clinoenstatite
cats = ca_tschermaks
jd = jadeite
mgc2 = hp_clinoenstatite
fec2 = hp_clinoferrosilite
hpcen = hp_clinoenstatite
hpcfs = hp_clinoferrosilite

# Perovskites
mgpv = mg_perovskite
mg_bridgmanite = mg_perovskite
fepv = fe_perovskite
fe_bridgmanite = fe_perovskite
alpv = al_perovskite
capv = ca_perovskite

# Ilmenite group
mgil = mg_akimotoite
feil = fe_akimotoite
co = corundum

# Garnet group
py = pyrope
al = almandine
gr = grossular
mgmj = mg_majorite
jdmj = jd_majorite

# Quartz polymorphs
qtz = quartz
coes = coesite
st = stishovite
seif = seifertite

# Post perovskites
mppv = mg_post_perovskite
fppv = fe_post_perovskite
appv = al_post_perovskite

# Magnesiowuestite
pe = periclase
wu = wuestite

# Calcium ferrite structured phases
mgcf = mg_ca_ferrite
fecf = fe_ca_ferrite
nacf = na_ca_ferrite

# Al2SiO5 polymorphs
ky = kyanite

# Nepheline group
neph = nepheline


# Solid solution aliases
c2c = c2c_pyroxene
cf = ca_ferrite_structured_phase
cpx = clinopyroxene
gt = garnet
il = akimotoite
ilmenite_group = akimotoite
mw = ferropericlase
magnesiowuestite = ferropericlase
ol = mg_fe_olivine
opx = orthopyroxene
plag = plagioclase
ppv = post_perovskite
pv = mg_fe_perovskite
mg_fe_bridgmanite = mg_fe_perovskite
mg_fe_silicate_perovskite = mg_fe_perovskite
ri = mg_fe_ringwoodite
spinel_group = mg_fe_aluminous_spinel
wa = mg_fe_wadsleyite
spinelloid_III = mg_fe_wadsleyite



from __future__ import absolute_import

# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.

import numpy as np
import scipy.optimize as opt
from . import equation_of_state as eos
from ..utils.math import bracket
import warnings


def bulk_modulus_fourth(volume, params):
    """
    compute the bulk modulus as per the fourth order
    birch-murnaghan equation of state.  Returns bulk
    modulus in the same units as the reference bulk
    modulus.  Pressure must be in :math:`[Pa]`.
    """

    invVrel = params["V_0"] / volume
    f = 0.5 * (pow(invVrel, 2.0 / 3.0) - 1.0)

    Xi = (3.0 / 4.0) * (4.0 - params["Kprime_0"])
    Zeta = (3.0 / 8.0) * (
        (params["K_0"] * params["Kprime_prime_0"])
        + params["Kprime_0"] * (params["Kprime_0"] - 7.0)
        + 143.0 / 9.0
    )

    K = (
        5.0
        * f
        * pow((1.0 + 2.0 * f), 5.0 / 2.0)
        * params["K_0"]
        * (1.0 - (2.0 * Xi * f) + (4.0 * Zeta * pow(f, 2.0)))
    ) + (
        pow(1.0 + (2.0 * f), 7.0 / 2.0)
        * params["K_0"]
        * (1.0 - (4.0 * Xi * f) + (12.0 * Zeta * pow(f, 2.0)))
    )

    return K


def volume_fourth_order(pressure, params):
    """
    Volume according to the fourth-order
    Birch-Murnaghan equation of state.

    :param pressure: Pressure in the same units that are supplied for the reference bulk
    modulus (params['K_0'])
    :type pressure: float
    :param params: parameter dictionary
    :type params: dictionary
    :return: V/V_0
    :rtype: float
    """

    def delta_pressure(x):
        return pressure_birch_murnaghan_fourth(params["V_0"] / x, params) - pressure

    try:
        sol = bracket(delta_pressure, params["V_0"], 1.0e-2 * params["V_0"])
    except ValueError:
        raise ValueError(
            "Cannot find a volume, perhaps you are outside of the range of validity for the equation of state?"
        )
    return opt.brentq(delta_pressure, sol[0], sol[1])


def pressure_birch_murnaghan_fourth(invVrel, params):
    """
    Pressure according to the fourth-order
    Birch-Murnaghan equation of state.

    :param invVrel: V/V_0
    :type invVrel: float or numpy array
    :param params: parameter dictionary
    :type params: dictionary
    :return: Pressure in the same units that are supplied for the reference bulk
    modulus (params['K_0'])
    :rtype: float or numpy array
    """

    f = 0.5 * (pow(invVrel, 2.0 / 3.0) - 1.0)
    Xi = (3.0 / 4.0) * (4.0 - params["Kprime_0"])
    Zeta = (3.0 / 8.0) * (
        (params["K_0"] * params["Kprime_prime_0"])
        + params["Kprime_0"] * (params["Kprime_0"] - 7.0)
        + 143.0 / 9.0
    )

    return (
        3.0
        * f
        * pow(1.0 + 2.0 * f, 5.0 / 2.0)
        * params["K_0"]
        * (1.0 - (2.0 * Xi * f) + (4.0 * Zeta * pow(f, 2.0)))
        + params["P_0"]
    )


class BM4(eos.EquationOfState):
    """
    Base class for the isothermal Birch Murnaghan equation of state.  This is fourth order in strain, and
    has no temperature dependence.
    """

    def volume(self, pressure, temperature, params):
        """
        Returns volume :math:`[m^3]` as a function of pressure :math:`[Pa]`.
        """
        return volume_fourth_order(pressure, params)

    def pressure(self, temperature, volume, params):
        return pressure_birch_murnaghan_fourth(params["V_0"] / volume, params)

    def isothermal_bulk_modulus_reuss(self, pressure, temperature, volume, params):
        """
        Returns isothermal bulk modulus :math:`K_T` :math:`[Pa]` as a function of pressure :math:`[Pa]`,
        temperature :math:`[K]` and volume :math:`[m^3]`.
        """
        return bulk_modulus_fourth(volume, params)

    def isentropic_bulk_modulus_reuss(self, pressure, temperature, volume, params):
        """
        Returns adiabatic bulk modulus :math:`K_s` of the mineral. :math:`[Pa]`.
        """
        return bulk_modulus_fourth(volume, params)

    def shear_modulus(self, pressure, temperature, volume, params):
        """
        Returns shear modulus :math:`G` of the mineral. :math:`[Pa]`
        """
        return 0.0

    def entropy(self, pressure, temperature, volume, params):
        """
        Returns the molar entropy :math:`\\mathcal{S}` of the mineral. :math:`[J/K/mol]`
        """
        return 0.0

    def molar_internal_energy(self, pressure, temperature, volume, params):
        """
        Returns the internal energy :math:`\\mathcal{E}` of the mineral. :math:`[J/mol]`
        """
        x = np.power(volume / params["V_0"], -1.0 / 3.0)
        x2 = x * x
        x4 = x2 * x2
        x6 = x4 * x2
        x8 = x4 * x4

        xi1 = 3.0 * (4.0 - params["Kprime_0"]) / 4.0
        xi2 = (
            3.0
            / 8.0
            * (
                params["K_0"] * params["Kprime_prime_0"]
                + params["Kprime_0"] * (params["Kprime_0"] - 7.0)
            )
            + 143.0 / 24.0
        )

        intPdV = (
            -9.0
            / 2.0
            * params["V_0"]
            * params["K_0"]
            * (
                (xi1 + 1.0) * (x4 / 4.0 - x2 / 2.0 + 1.0 / 4.0)
                - xi1 * (x6 / 6.0 - x4 / 4.0 + 1.0 / 12.0)
                + xi2 * (x8 / 8 - x6 / 2 + 3.0 * x4 / 4.0 - x2 / 2.0 + 1.0 / 8.0)
            )
        )

        return -intPdV + params["E_0"]

    def gibbs_free_energy(self, pressure, temperature, volume, params):
        """
        Returns the Gibbs free energy :math:`\\mathcal{G}` of the mineral. :math:`[J/mol]`
        """
        # G = int VdP = [PV] - int PdV = E + PV

        return (
            self.molar_internal_energy(pressure, temperature, volume, params)
            + volume * pressure
        )

    def molar_heat_capacity_v(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return a very large number. :math:`[J/K/mol]`
        """
        return 1.0e99

    def molar_heat_capacity_p(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return a very large number. :math:`[J/K/mol]`
        """
        return 1.0e99

    def thermal_expansivity(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return zero. :math:`[1/K]`
        """
        return 0.0

    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return zero. :math:`[unitless]`
        """
        return 0.0

    def validate_parameters(self, params):
        """
        Check for existence and validity of the parameters
        """

        if "E_0" not in params:
            params["E_0"] = 0.0
        if "P_0" not in params:
            params["P_0"] = 0.0

        # If G and Gprime are not included this is presumably deliberate,
        # as we can model density and bulk modulus just fine without them,
        # so just add them to the dictionary as nans
        if "G_0" not in params:
            params["G_0"] = float("nan")
        if "Gprime_0" not in params:
            params["Gprime_0"] = float("nan")

        # Check that all the required keys are in the dictionary
        expected_keys = ["V_0", "K_0", "Kprime_0"]
        for k in expected_keys:
            if k not in params:
                raise KeyError("params object missing parameter : " + k)

        # Finally, check that the values are reasonable.
        if params["P_0"] < 0.0:
            warnings.warn("Unusual value for P_0", stacklevel=2)
        if params["V_0"] < 1.0e-7 or params["V_0"] > 1.0e-3:
            warnings.warn("Unusual value for V_0", stacklevel=2)
        if params["K_0"] < 1.0e9 or params["K_0"] > 1.0e13:
            warnings.warn("Unusual value for K_0", stacklevel=2)
        if params["Kprime_0"] < 0.0 or params["Kprime_0"] > 10.0:
            warnings.warn("Unusual value for Kprime_0", stacklevel=2)
        if params["Kprime_prime_0"] > 0.0 or params["Kprime_prime_0"] < -10.0:
            warnings.warn("Unusual value for Kprime_prime_0", stacklevel=2)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2015 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import

from os import path
import numpy as np
import scipy.optimize as opt

try:
    # scipy's factorial was moved to special in scipy 1.3.0+
    from scipy.special import factorial
except ImportError:
    from scipy.misc import factorial

import warnings

from . import equation_of_state as eos
from .. import constants as constants
from ..utils.chemistry import read_masses
from ..utils.math import bracket

atomic_masses = read_masses()
# energy_states should provide the energies and degeneracies of each electronic level in a variety of elements


class DKS_L(eos.EquationOfState):
    """
    Base class for the finite strain liquid equation of state detailed
    in :cite:`deKoker2013` (supplementary materials).
    """

    """
    Ideal gas contributions (translational and electronic)
    to thermodynamic properties
    """

    def _ln_partition_function(self, mass, temperature):
        """
        Calculates the natural log of the partition function
        """
        return 3.0 / 2.0 * np.log(temperature) + 3.0 / 2.0 * np.log(
            mass * constants.Boltzmann / (2 * np.pi * constants.Dirac * constants.Dirac)
        )

    def _F_ig(self, temperature, volume, params):
        """
        The ideal gas contribution to the helmholtz free energy
        Eq. S6, see also eq. 16.72 of Callen., 1985; p. 373
        """

        V = volume / constants.Avogadro
        figoverRT = 0.0
        for element, N in params["formula"].items():  # N is a.p.f.u
            if N > 1.0e-5:
                mass = atomic_masses[element] / constants.Avogadro
                figoverRT += -N * (
                    np.log(V) + self._ln_partition_function(mass, temperature) + 1.0
                ) + N * np.log(N)
        return constants.gas_constant * temperature * figoverRT

    def _S_ig(self, temperature, volume, params):
        """
        The ideal gas contribution to the entropy
        """

        V = volume / constants.Avogadro
        entropy_sum = 0.0
        for element, N in params["formula"].items():  # N is a.p.f.u
            if N > 1.0e-5:
                mass = atomic_masses[element] / constants.Avogadro
                entropy_sum -= -N * (
                    np.log(V)
                    + self._ln_partition_function(mass, temperature)
                    + 5.0 / 2.0
                ) + N * np.log(N)
        return constants.gas_constant * entropy_sum

    def _C_v_ig(self, temperature, volume, params):
        """
        The ideal gas contribution to the heat capacity
        """

        n_atoms = 0
        for element, N in params["formula"].items():
            n_atoms += N
        return 1.5 * constants.gas_constant * n_atoms

    def _P_ig(self, temperature, volume, params):
        """
        The ideal gas contribution to the pressure
        PV = nRT
        """

        n_atoms = 0
        for element, N in params["formula"].items():
            n_atoms += N
        return n_atoms * constants.gas_constant * temperature / volume

    def _K_T_ig(self, temperature, volume, params):
        """
        The ideal gas contribution to the isothermal bulk modulus
        V * d/dV(-nRT/V) = V*nRT/V^2
        """
        n_atoms = 0
        for element, N in params["formula"].items():
            n_atoms += N
        return n_atoms * constants.gas_constant * temperature / volume

    def _alphaK_T_ig(self, temperature, volume, params):
        """
        The ideal gas contribution to the product of the
        thermal expansivity and isothermal bulk modulus
        d/dT(nRT/V) = nR/V
        """

        n_atoms = 0
        for element, N in params["formula"].items():
            n_atoms += N
        return n_atoms * constants.gas_constant / volume

    """
    Electronic contributions to thermodynamic properties
    """

    def _zeta(
        self, temperature, volume, params
    ):  # eq. S5a, beta in deKoker thesis (3.34)
        return params["zeta_0"] * (np.power(volume / params["el_V_0"], params["xi"]))

    def _dzetadV(self, temperature, volume, params):
        return (
            params["zeta_0"]
            * params["xi"]
            * (np.power(volume / params["el_V_0"], params["xi"]))
            / volume
        )

    def _d2zetadV2(self, temperature, volume, params):
        return (
            params["zeta_0"]
            * params["xi"]
            * (params["xi"] - 1.0)
            * (np.power(volume / params["el_V_0"], params["xi"]))
            / volume
            / volume
        )

    def _Tel(self, temperature, volume, params):  # eq. S5b
        return params["Tel_0"] * (np.power(volume / params["el_V_0"], params["eta"]))

    def _dTeldV(self, temperature, volume, params):
        return (
            params["Tel_0"]
            * params["eta"]
            * (np.power(volume / params["el_V_0"], params["eta"]))
            / volume
        )

    def _d2TeldV2(self, temperature, volume, params):
        return (
            params["Tel_0"]
            * params["eta"]
            * (params["eta"] - 1.0)
            * (np.power(volume / params["el_V_0"], params["eta"]))
            / volume
            / volume
        )

    def _gimel(
        self, temperature_el, temperature, volume, params
    ):  # -F_el/zeta, 3.30 in de Koker thesis
        return 0.5 * (
            temperature * temperature - temperature_el * temperature_el
        ) - temperature * temperature_el * np.log(temperature / temperature_el)

    def _dgimeldTel(self, temperature_el, temperature, volume, params):
        return (temperature - temperature_el) - temperature * np.log(
            temperature / temperature_el
        )

    def _dgimeldT(self, temperature_el, temperature, volume, params):
        return (temperature - temperature_el) - temperature_el * np.log(
            temperature / temperature_el
        )

    def _d2gimeldTdTel(self, temperature_el, temperature, volume, params):
        return -np.log(temperature / temperature_el)

    def _d2gimeldTel2(self, temperature_el, temperature, volume, params):
        return (temperature / temperature_el) - 1.0

    def _F_el(self, temperature, volume, params):  # F_el
        temperature_el = self._Tel(temperature, volume, params)
        if temperature < temperature_el:
            F_el = 0
        else:
            F_el = -self._zeta(temperature, volume, params) * self._gimel(
                temperature_el, temperature, volume, params
            )
        return F_el

    def _S_el(self, temperature, volume, params):  # S_el
        temperature_el = self._Tel(temperature, volume, params)
        if temperature < temperature_el:
            S_el = 0
        else:
            S_el = self._zeta(temperature, volume, params) * self._dgimeldT(
                temperature_el, temperature, volume, params
            )
        return S_el

    def _P_el(self, temperature, volume, params):  # P_el
        temperature_el = self._Tel(temperature, volume, params)
        if temperature < temperature_el:
            P_el = 0
        else:
            P_el = self._dzetadV(temperature, volume, params) * self._gimel(
                temperature_el, temperature, volume, params
            ) + self._zeta(temperature, volume, params) * self._dTeldV(
                temperature, volume, params
            ) * self._dgimeldTel(
                temperature_el, temperature, volume, params
            )
        return P_el

    def _K_T_el(self, temperature, volume, params):  # K_T_el
        temperature_el = self._Tel(temperature, volume, params)
        if temperature < temperature_el:
            K_T_el = 0
        else:
            K_T_el = -volume * (
                self._d2zetadV2(temperature, volume, params)
                * self._gimel(temperature_el, temperature, volume, params)
                + 2.0
                * self._dzetadV(temperature, volume, params)
                * self._dgimeldTel(temperature_el, temperature, volume, params)
                * self._dTeldV(temperature, volume, params)
                + self._zeta(temperature, volume, params)
                * (
                    self._d2TeldV2(temperature, volume, params)
                    * self._dgimeldTel(temperature_el, temperature, volume, params)
                    + self._dTeldV(temperature, volume, params)
                    * self._dTeldV(temperature, volume, params)
                    * self._d2gimeldTel2(temperature_el, temperature, volume, params)
                )
            )
        return K_T_el

    def _alphaK_T_el(self, temperature, volume, params):  # (alphaK_T)_el
        temperature_el = self._Tel(temperature, volume, params)
        if temperature < temperature_el:
            alphaK_T_el = 0
        else:
            alphaK_T_el = self._dzetadV(temperature, volume, params) * self._dgimeldT(
                temperature_el, temperature, volume, params
            ) + self._zeta(temperature, volume, params) * self._d2gimeldTdTel(
                temperature_el, temperature, volume, params
            ) * self._dTeldV(
                temperature, volume, params
            )
        return alphaK_T_el

    def _C_v_el(self, temperature, volume, params):  # C_el, eq. 3.28 of de Koker thesis
        temperature_el = self._Tel(temperature, volume, params)
        zeta = self._zeta(temperature, volume, params)

        if temperature > temperature_el:
            Cv_el = zeta * (temperature - temperature_el)
        else:
            Cv_el = 0.0
        return Cv_el

    """
    Excess (bonding) contributions to thermodynamic properties
    """

    # Finite strain
    def _finite_strain(self, temperature, volume, params):  # f(V), eq. S3a
        return (1.0 / 2.0) * (np.power(params["V_0"] / volume, 2.0 / 3.0) - 1.0)

    def _dfdV(self, temperature, volume, params):  # f(V), eq. S3a
        return (-1.0 / 3.0) * np.power(params["V_0"] / volume, 2.0 / 3.0) / volume

    def _d2fdV2(self, temperature, volume, params):
        return (
            (5.0 / 9.0) * np.power(params["V_0"] / volume, 2.0 / 3.0) / volume / volume
        )

    # Temperature
    def _theta(self, temperature, volume, params):  # theta, eq. S3b
        return np.power(temperature / params["T_0"], params["m"]) - 1.0

    def _dthetadT(self, temperature, volume, params):
        return (
            params["m"]
            * np.power(temperature / params["T_0"], params["m"])
            / temperature
        )

    def _d2thetadT2(self, temperature, volume, params):
        return (
            params["m"]
            * (params["m"] - 1.0)
            * np.power(temperature / params["T_0"], params["m"])
            / temperature
            / temperature
        )

    def _F_xs(self, temperature, volume, params):  # F_xs, eq. S2
        f = self._finite_strain(temperature, volume, params)
        theta = self._theta(temperature, volume, params)
        energy = 0.0
        for i in range(len(params["a"])):
            ifact = factorial(i, exact=False)
            for j in range(len(params["a"][0])):
                jfact = factorial(j, exact=False)
                energy += (
                    params["a"][i][j]
                    * np.power(f, i)
                    * np.power(theta, j)
                    / ifact
                    / jfact
                )
        return energy

    def _S_xs(self, temperature, volume, params):  # F_xs, eq. 3.18
        f = self._finite_strain(temperature, volume, params)
        theta = self._theta(temperature, volume, params)
        entropy = 0.0
        for i in range(len(params["a"])):
            ifact = factorial(i, exact=False)
            for j in range(len(params["a"][0])):
                if j > 0:
                    jfact = factorial(j, exact=False)
                    entropy += (
                        j
                        * params["a"][i][j]
                        * np.power(f, i)
                        * np.power(theta, j - 1.0)
                        / ifact
                        / jfact
                    )
        return -self._dthetadT(temperature, volume, params) * entropy

    def _P_xs(self, temperature, volume, params):  # P_xs, eq. 3.17 of de Koker thesis
        f = self._finite_strain(temperature, volume, params)
        theta = self._theta(temperature, volume, params)
        pressure = 0.0
        for i in range(len(params["a"])):
            ifact = factorial(i, exact=False)
            if i > 0:
                for j in range(len(params["a"][0])):
                    jfact = factorial(j, exact=False)
                    pressure += (
                        float(i)
                        * params["a"][i][j]
                        * np.power(f, float(i) - 1.0)
                        * np.power(theta, float(j))
                        / ifact
                        / jfact
                    )
        return -self._dfdV(temperature, volume, params) * pressure

    def _K_T_xs(
        self, temperature, volume, params
    ):  # K_T_xs, eq. 3.20 of de Koker thesis
        f = self._finite_strain(temperature, volume, params)
        theta = self._theta(temperature, volume, params)
        K_ToverV = 0.0
        for i in range(len(params["a"])):
            ifact = factorial(i, exact=False)
            for j in range(len(params["a"][0])):
                if i > 0:
                    jfact = factorial(j, exact=False)
                    prefactor = (
                        float(i)
                        * params["a"][i][j]
                        * np.power(theta, float(j))
                        / ifact
                        / jfact
                    )
                    K_ToverV += (
                        prefactor
                        * self._d2fdV2(temperature, volume, params)
                        * np.power(f, float(i - 1))
                    )
                if i > 1:
                    dfdV = self._dfdV(temperature, volume, params)
                    K_ToverV += (
                        prefactor
                        * dfdV
                        * dfdV
                        * float(i - 1)
                        * np.power(f, float(i - 2))
                    )
        return volume * K_ToverV

    def _alphaK_T_xs(self, temperature, volume, params):  # eq. 3.21 of de Koker thesis
        f = self._finite_strain(temperature, volume, params)
        theta = self._theta(temperature, volume, params)
        sum_factors = 0.0
        for i in range(len(params["a"])):
            ifact = factorial(i, exact=False)
            if i > 0:
                for j in range(len(params["a"][0])):
                    if j > 0:
                        jfact = factorial(j, exact=False)
                        sum_factors += (
                            float(i)
                            * float(j)
                            * params["a"][i][j]
                            * np.power(f, float(i - 1))
                            * np.power(theta, float(j - 1))
                            / ifact
                            / jfact
                        )

        return (
            -self._dfdV(temperature, volume, params)
            * self._dthetadT(temperature, volume, params)
            * sum_factors
        )

    def _C_v_xs(
        self, temperature, volume, params
    ):  # Cv_xs, eq. 3.22 of de Koker thesis
        f = self._finite_strain(temperature, volume, params)
        theta = self._theta(temperature, volume, params)
        C_voverT = 0.0
        for i in range(len(params["a"])):
            ifact = factorial(i, exact=False)
            for j in range(len(params["a"][0])):
                if j > 0:
                    jfact = factorial(j, exact=False)
                    prefactor = (
                        float(j)
                        * params["a"][i][j]
                        * np.power(f, float(i))
                        / ifact
                        / jfact
                    )
                    C_voverT += (
                        prefactor
                        * self._d2thetadT2(temperature, volume, params)
                        * np.power(theta, float(j - 1))
                    )
                if j > 1:
                    dthetadT = self._dthetadT(temperature, volume, params)
                    C_voverT += (
                        prefactor
                        * dthetadT
                        * dthetadT
                        * float(j - 1)
                        * np.power(theta, float(j - 2))
                    )
        return -temperature * C_voverT

    """
    Magnetic contributions to thermodynamic properties
    (as found in Ramo and Stixrude, 2014)
    """

    def _spin(self, temperature, volume, params):
        S_a = 0.0
        S_b = 0.0
        numerator = 0.0
        numerator_2 = 0.0
        n_atoms = 0.0
        if "spin_a" in params:
            for element, N in params["formula"].items():
                if element == "Fe":
                    n_atoms += N

            VoverVx = volume / params["V_0"]
            S_a = params["spin_a"][0] + params["spin_a"][1] * VoverVx
            S_b = (
                params["spin_b"][0]
                + params["spin_b"][1] / VoverVx
                + params["spin_b"][2] / (np.power(VoverVx, 2.0))
                + params["spin_b"][3] / (np.power(VoverVx, 3.0))
            )

            # S = S_a*T + S_b
            # d(2S + 1)/dV
            numerator = (
                -2.0
                * (
                    -params["spin_a"][1] * temperature
                    + params["spin_b"][1] / (np.power(VoverVx, 2.0))
                    + 2.0 * params["spin_b"][2] / (np.power(VoverVx, 3.0))
                    + 3.0 * params["spin_b"][3] / (np.power(VoverVx, 4.0))
                )
                / params["V_0"]
            )

            # d2S/dV2
            numerator_2 = 2.0 * (
                (
                    2.0 * params["spin_b"][1] / (np.power(VoverVx, 3.0))
                    + 6.0 * params["spin_b"][2] / (np.power(VoverVx, 4.0))
                    + 12.0 * params["spin_b"][3] / (np.power(VoverVx, 5.0))
                )
                / np.power(params["V_0"], 2.0)
            )
        return S_a, S_b, numerator, numerator_2, n_atoms

    def _F_mag(self, temperature, volume, params):
        S_a, S_b, numerator, numerator_2, n_atoms = self._spin(
            temperature, volume, params
        )
        S = S_a * temperature + S_b
        return -n_atoms * constants.gas_constant * temperature * np.log(2.0 * S + 1.0)

    def _S_mag(self, temperature, volume, params):
        S_a, S_b, numerator, numerator_2, n_atoms = self._spin(
            temperature, volume, params
        )
        S = S_a * temperature + S_b
        return (
            n_atoms
            * constants.gas_constant
            * ((2.0 * S_a * temperature / (2.0 * S + 1.0) + np.log(2.0 * S + 1.0)))
        )

    def _P_mag(self, temperature, volume, params):
        S_a, S_b, numerator, numerator_2, n_atoms = self._spin(
            temperature, volume, params
        )
        S = S_a * temperature + S_b
        dFdV = (
            -n_atoms
            * constants.gas_constant
            * temperature
            * numerator
            / (2.0 * S + 1.0)
        )
        return -dFdV

    def _K_T_mag(self, temperature, volume, params):
        S_a, S_b, numerator, numerator_2, n_atoms = self._spin(
            temperature, volume, params
        )
        S = S_a * temperature + S_b
        dFdV = numerator / (2.0 * S + 1.0)
        d2FdV2 = numerator_2 / (2.0 * S + 1.0) - np.power(dFdV, 2.0)

        return -volume * n_atoms * constants.gas_constant * temperature * d2FdV2

    def _alphaK_T_mag(
        self, temperature, volume, params
    ):  # WARNING: numeric differentiation a.t.m.
        return self._P_mag(temperature + 0.5, volume, params) - self._P_mag(
            temperature - 0.5, volume, params
        )

    def _C_v_mag(self, temperature, volume, params):
        S_a, S_b, numerator, numerator_2, n_atoms = self._spin(
            temperature, volume, params
        )
        S = S_a * temperature + S_b
        return (
            n_atoms
            * constants.gas_constant
            * temperature
            * 4.0
            * S_a
            * (S_a * temperature + 2.0 * S_b + 1.0)
            / np.power(2.0 * S + 1.0, 2.0)
        )

    def _aK_T(self, temperature, volume, params):
        aK_T = (
            self._alphaK_T_ig(temperature, volume, params)
            + self._alphaK_T_el(temperature, volume, params)
            + self._alphaK_T_xs(temperature, volume, params)
            + self._alphaK_T_mag(temperature, volume, params)
        )
        return aK_T

    # Pressure
    def pressure(self, temperature, volume, params):
        P = (
            self._P_ig(temperature, volume, params)
            + self._P_el(temperature, volume, params)
            + self._P_xs(temperature, volume, params)
            + self._P_mag(temperature, volume, params)
        )
        return P

    def volume(self, pressure, temperature, params):
        _delta_pressure = (
            lambda x, pressure, temperature, params: pressure
            - self.pressure(temperature, x, params)
        )

        # we need to have a sign change in [a,b] to find a zero. Let us start with a
        # conservative guess:
        args = (pressure, temperature, params)
        try:
            sol = bracket(_delta_pressure, params["V_0"], 1.0e-2 * params["V_0"], args)
        except ValueError:
            raise Exception(
                "Cannot find a volume, perhaps you are outside of the range of validity for the equation of state?"
            )
        return opt.brentq(_delta_pressure, sol[0], sol[1], args=args)

    def isothermal_bulk_modulus_reuss(self, pressure, temperature, volume, params):
        """
        Returns isothermal bulk modulus :math:`[Pa]`
        """
        K_T = (
            self._K_T_ig(temperature, volume, params)
            + self._K_T_el(temperature, volume, params)
            + self._K_T_xs(temperature, volume, params)
            + self._K_T_mag(temperature, volume, params)
        )
        return K_T

    def isentropic_bulk_modulus_reuss(self, pressure, temperature, volume, params):
        """
        Returns adiabatic bulk modulus. :math:`[Pa]`
        """
        K_S = self.isothermal_bulk_modulus_reuss(
            pressure, temperature, volume, params
        ) * (
            1.0
            + temperature
            * self.thermal_expansivity(pressure, temperature, volume, params)
            * self.grueneisen_parameter(pressure, temperature, volume, params)
        )
        return K_S

    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        Returns grueneisen parameter. :math:`[unitless]`
        """
        gamma = (
            self._aK_T(temperature, volume, params)
            * volume
            / self.molar_heat_capacity_v(pressure, temperature, volume, params)
        )
        return gamma

    def shear_modulus(self, pressure, temperature, volume, params):
        """
        Returns shear modulus. :math:`[Pa]`
        Zero for fluids
        """
        return 0.0

    def molar_heat_capacity_v(self, pressure, temperature, volume, params):
        """
        Returns heat capacity at constant volume. :math:`[J/K/mol]`
        """
        C_v = (
            self._C_v_ig(temperature, volume, params)
            + self._C_v_el(temperature, volume, params)
            + self._C_v_xs(temperature, volume, params)
            + self._C_v_mag(temperature, volume, params)
        )
        return C_v

    def molar_heat_capacity_p(self, pressure, temperature, volume, params):
        """
        Returns heat capacity at constant pressure. :math:`[J/K/mol]`
        """
        C_p = self.molar_heat_capacity_v(pressure, temperature, volume, params) * (
            1.0
            + temperature
            * self.thermal_expansivity(pressure, temperature, volume, params)
            * self.grueneisen_parameter(pressure, temperature, volume, params)
        )
        return C_p

    def thermal_expansivity(self, pressure, temperature, volume, params):
        """
        Returns thermal expansivity. :math:`[1/K]`
        """
        alpha = self._aK_T(
            temperature, volume, params
        ) / self.isothermal_bulk_modulus_reuss(0.0, temperature, volume, params)
        return alpha

    def gibbs_free_energy(self, pressure, temperature, volume, params):
        """
        Returns the Gibbs free energy at the pressure and temperature of the mineral [J/mol]
        """
        G = (
            self.helmholtz_free_energy(pressure, temperature, volume, params)
            + pressure * volume
        )
        return G

    def entropy(self, pressure, temperature, volume, params):
        """
        Returns the entropy at the pressure and temperature of the mineral [J/K/mol]
        """
        S = (
            self._S_ig(temperature, volume, params)
            + self._S_el(temperature, volume, params)
            + self._S_xs(temperature, volume, params)
            + self._S_mag(temperature, volume, params)
        )
        return S

    def enthalpy(self, pressure, temperature, volume, params):
        """
        Returns the enthalpy at the pressure and temperature of the mineral [J/mol]
        """
        H = (
            self.helmholtz_free_energy(pressure, temperature, volume, params)
            + temperature * self.entropy(pressure, temperature, volume, params)
            + pressure * self.volume(pressure, temperature, params)
        )
        return H

    def helmholtz_free_energy(self, pressure, temperature, volume, params):
        """
        Returns the Helmholtz free energy at the pressure and temperature of the mineral [J/mol]
        """
        F = (
            self._F_ig(temperature, volume, params)
            + self._F_el(temperature, volume, params)
            + self._F_xs(temperature, volume, params)
            + self._F_mag(temperature, volume, params)
        )
        return F

    def molar_internal_energy(self, pressure, temperature, volume, params):
        E = self.helmholtz_free_energy(
            pressure, temperature, volume, params
        ) + temperature * self.entropy(pressure, temperature, volume, params)
        return E

    def validate_parameters(self, params):
        """
        Check for existence and validity of the parameters
        """

        # Check that all the required keys are in the dictionary
        expected_keys = [
            "V_0",
            "T_0",
            "O_theta",
            "O_f",
            "m",
            "a",
            "zeta_0",
            "xi",
            "Tel_0",
            "eta",
        ]
        for k in expected_keys:
            if k not in params:
                raise KeyError("params object missing parameter : " + k)

        # Sometimes the standard electronic volume is different to V_0.
        # If not, make it the same.
        if "el_V_0" not in params:
            params["el_V_0"] = params["V_0"]
